import numbers

import numpy as np

from ..problem import Problem
from .metropolis import MetropolisSampler


class AdaptiveMetropolisSampler(MetropolisSampler):
    """Metropolis-Hastings sampler with adaptive proposal covariance.

    A core problem of the standard Metropolis-Hastings sampler is
    its fixed proposal distribution, which must be manually tuned.
    This sampler adapts the proposal distribution during the sampling
    process based on previous samples.
    It adapts the correlation structure and the scaling factor of the
    proposal distribution.
    For both parts, there exist a variety of methods, see

    * Ballnus et al. 2017.
      Comprehensive benchmarking of Markov chain Monte Carlo methods for
      dynamical systems
      (https://doi.org/10.1186/s12918-017-0433-1)
    * Andrieu et al. 2008.
      A tutorial on adaptive MCMC
      (https://doi.org/10.1007/s11222-008-9110-y)

    for a review.

    Here, we approximate the covariance matrix via a weighted average of
    current and earlier samples,
    with a decay factor determining the relative contribution of the
    current sample and earlier ones to the weighted average of mean and
    covariance.
    The scaling factor we aim to converge to a fixed target acceptance rate
    of 0.234, as suggested by theoretical results.
    The implementation is based on:

    * Lacki et al. 2015.
      State-dependent swap strategies and automatic reduction of number of
      temperatures in adaptive parallel tempering algorithm
      (https://doi.org/10.1007/s11222-015-9579-0)
    * Miasojedow et al. 2013.
      An adaptive parallel tempering algorithm
      (https://doi.org/10.1080/10618600.2013.778779)

    In turn, these are based on adaptive MCMC as discussed in:

    * Haario et al. 2001.
      An adaptive Metropolis algorithm
      (https://doi.org/10.2307/3318737)

    For reference matlab implementations see:

    * https://github.com/ICB-DCM/PESTO/blob/master/private/performPT.m
    * https://github.com/ICB-DCM/PESTO/blob/master/private/updateStatistics.m
    """

    def __init__(self, options: dict = None):
        super().__init__(options)
        self._cov = None
        self._mean_hist = None
        self._cov_hist = None
        self._cov_scale = None

    @classmethod
    def default_options(cls):
        """Return the default options for the sampler."""
        return {
            # controls adaptation degeneration velocity of the proposals
            # in [0, 1], with 0 -> no adaptation, i.e. classical
            # Metropolis-Hastings
            "decay_constant": 0.51,
            # number of samples before adaptation decreases significantly.
            # a higher value reduces the impact of early adaptation
            "threshold_sample": 1,
            # regularization factor for ill-conditioned cov matrices of
            # the adapted proposal density. regularization might happen if the
            # eigenvalues of the cov matrix strongly differ in order
            # of magnitude. in this case, the algorithm adds a small
            # diag matrix to the cov matrix with elements of this factor
            "reg_factor": 1e-6,
            # initial covariance matrix. defaults to a unit matrix
            "cov0": None,
            # target acceptance rate
            "target_acceptance_rate": 0.234,
            # show progress
            "show_progress": None,
        }

    def initialize(self, problem: Problem, x0: np.ndarray):
        """Initialize the sampler."""
        super().initialize(problem, x0)

        if self.options["cov0"] is not None:
            cov0 = self.options["cov0"]
            if isinstance(cov0, numbers.Real):
                cov0 = float(cov0) * np.eye(len(x0))
        else:
            cov0 = np.eye(len(x0))
        self._cov = regularize_covariance(cov0, self.options["reg_factor"])
        self._mean_hist = self.trace_x[-1]
        self._cov_hist = self._cov
        self._cov_scale = 1.0

    def _propose_parameter(self, x: np.ndarray):
        x_new = np.random.multivariate_normal(x, self._cov)
        return x_new

    def _update_proposal(
        self, x: np.ndarray, lpost: float, log_p_acc: float, n_sample_cur: int
    ):
        # parse options
        decay_constant = self.options["decay_constant"]
        threshold_sample = self.options["threshold_sample"]
        reg_factor = self.options["reg_factor"]
        target_acceptance_rate = self.options["target_acceptance_rate"]

        # compute historical mean and covariance
        self._mean_hist, self._cov_hist = update_history_statistics(
            mean=self._mean_hist,
            cov=self._cov_hist,
            x_new=x,
            n_cur_sample=max(n_sample_cur + 1, threshold_sample),
            decay_constant=decay_constant,
        )

        # compute covariance scaling factor based on the target acceptance rate
        self._cov_scale *= np.exp(
            (np.exp(log_p_acc) - target_acceptance_rate)
            / np.power(n_sample_cur + 1, decay_constant)
        )

        # set proposal covariance
        # TODO check publication
        self._cov = self._cov_scale * self._cov_hist

        # regularize proposal covariance
        self._cov = regularize_covariance(cov=self._cov, reg_factor=reg_factor)


def update_history_statistics(
    mean: np.ndarray,
    cov: np.ndarray,
    x_new: np.ndarray,
    n_cur_sample: int,
    decay_constant: float,
) -> tuple[np.ndarray, np.ndarray]:
    """Update sampling mean and covariance matrix via weighted average.

    Update sampling mean and covariance matrix based on the previous
    estimate and the most recent sample via a weighted average,
    with gradually decaying update rate.

    Parameters
    ----------
    mean:
        The estimated mean of the samples, that was calculated in the previous
        iteration.
    cov:
        The estimated covariance matrix of the sample, that was calculated in
        the previous iteration.
    x_new:
        Most recent sample.
    n_cur_sample:
        Current number of samples.
    decay_constant:
        Adaption decay, in [0, 1]. Higher values result in faster decays, such
        that later iterations influence the adaption more weakly.

    Returns
    -------
    mean, cov:
        The updated values for the estimated mean and the estimated covariance
        matrix of the sample.
    """
    update_rate = n_cur_sample ** (-decay_constant)

    mean = (1 - update_rate) * mean + update_rate * x_new

    dx = x_new - mean
    cov = (1 - update_rate) * cov + update_rate * dx.reshape(
        (-1, 1)
    ) @ dx.reshape((1, -1))

    return mean, cov


def regularize_covariance(cov: np.ndarray, reg_factor: float) -> np.ndarray:
    """
    Regularize the estimated covariance matrix of the sample.

    Useful if the estimated covariance matrix is ill-conditioned.
    Increments the diagonal a little to ensure positivity.

    Parameters
    ----------
    cov:
        Estimate of the covariance matrix of the sample.
    reg_factor:
        Regularization factor. Larger values result in stronger regularization.

    Returns
    -------
    cov:
        Regularized estimate of the covariance matrix of the sample.
    """
    eig = np.linalg.eigvals(cov)
    eig_min = min(eig)
    if eig_min <= 0:
        cov += (abs(eig_min) + reg_factor) * np.eye(cov.shape[0])
    return cov



"""Engines without parallelization."""

from typing import Any

from ..util import tqdm
from .base import Engine
from .task import Task


class SingleCoreEngine(Engine):
    """
    Dummy engine for sequential execution on one core.

    .. note:: The objective itself may be multithreaded.
    """

    def __init__(self):
        super().__init__()

    def execute(
        self, tasks: list[Task], progress_bar: bool = None
    ) -> list[Any]:
        """Execute all tasks in a simple for loop sequentially.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.

        Returns
        -------
        A list of results.
        """
        results = []
        for task in tqdm(
            tasks,
            enable=progress_bar,
        ):
            results.append(task.execute())

        return results



"""Enhanced Scatter Search.

See papers on ESS :footcite:p:`EgeaBal2009,EgeaMar2010`,
CESS :footcite:p:`VillaverdeEge2012`, and saCeSS :footcite:p:`PenasGon2017`.
"""

from __future__ import annotations

import enum
import logging
import time
from typing import Protocol
from warnings import warn

import numpy as np

import pypesto.optimize
from pypesto import OptimizerResult, Problem
from pypesto.history import MemoryHistory
from pypesto.startpoint import StartpointMethod

from .function_evaluator import FunctionEvaluator, create_function_evaluator
from .refset import RefSet

logger = logging.getLogger(__name__)

__all__ = ["ESSOptimizer", "ESSExitFlag"]


class ESSExitFlag(int, enum.Enum):
    """Scatter search exit flags.

    Exit flags used by :class:`pypesto.ess.ESSOptimizer` and
    :class:`pypesto.ess.SacessOptimizer`.
    """

    # ESS did not run/finish yet
    DID_NOT_RUN = 0
    # Exited after reaching the maximum number of iterations
    MAX_ITER = -1
    # Exited after exhausting function evaluation budget
    MAX_EVAL = -2
    # Exited after exhausting wall-time budget
    MAX_TIME = -3
    # Termination because of other reasons than exit criteria
    ERROR = -99


class OptimizerFactory(Protocol):
    def __call__(
        self, max_eval: float, max_walltime_s: float
    ) -> pypesto.optimize.Optimizer:
        """Create a new optimizer instance.

        Parameters
        ----------
        max_eval:
            Maximum number of objective functions allowed.
        max_walltime_s:
            Maximum walltime in seconds.
        """
        ...


class ESSOptimizer:
    """Enhanced Scatter Search (ESS) global optimization.

    Scatter search is a meta-heuristic for global optimization. A set of points
    (the reference set, RefSet) is iteratively adapted to explore the parameter
    space and to follow promising directions.

    This implementation is based on :footcite:p:`EgeaBal2009,EgeaMar2010`,
    but does not implement any constraint handling beyond box constraints.

    The basic steps of ESS are:

    * Initialization: Generate a diverse set of points (RefSet) in the
      parameter space.
    * Recombination: Generate new points by recombining the RefSet points.
    * Improvement: Improve the RefSet by replacing points with better ones.

    The steps are repeated until a stopping criterion is met.

    ESS is gradient-free, unless a gradient-based local optimizer is used
    (``local_optimizer``).

    Hyperparameters
    ---------------

    Various hyperparameters control the behavior of ESS.
    Initialization is controlled by ``dim_refset`` and ``n_diverse``.
    Local optimizations are controlled by ``local_optimizer``, ``local_n1``,
    ``local_n2``, and ``balance``.

    Exit criteria
    -------------

    The optimization stops if any of the following criteria are met:

    * The maximum number of iterations is reached (``max_iter``).
    * The maximum number of objective function evaluations is reached
      (``max_eval``).
    * The maximum wall-time is reached (``max_walltime_s``).

    One of these criteria needs to be provided.
    Note that the wall-time and function evaluation criteria are not checked
    after every single function evaluation, and thus, the actual number of
    function evaluations may slightly exceed the given value.

    Parallelization
    ---------------

    Objective function evaluations inside :class:`ESSOptimizer` can be
    parallelized using multiprocessing or multithreading by passing a value
    >1 for ``n_procs`` or ``n_threads``, respectively.


    .. seealso::

       :class:`pypesto.optimize.ess.sacess.SacessOptimizer`

    .. footbibliography::
    """

    def __init__(
        self,
        *,
        max_iter: int = None,
        dim_refset: int = None,
        local_n1: int = 1,
        local_n2: int = 10,
        balance: float = 0.5,
        local_optimizer: pypesto.optimize.Optimizer
        | OptimizerFactory
        | None = None,
        max_eval=None,
        n_diverse: int = None,
        n_procs=None,
        n_threads=None,
        max_walltime_s=None,
        result_includes_refset: bool = False,
    ):
        r"""Construct new ESS instance.

        For plausible values of hyperparameters, see :footcite:t:`VillaverdeEge2012`.

        Parameters
        ----------
        dim_refset:
            Size of the ReferenceSet. Note that in every iteration at least
            ``dim_refset**2 - dim_refset`` function evaluations will occur.
        max_iter:
            Maximum number of ESS iterations.
        local_n1:
            Minimum number of iterations before first local search.
            Ignored if ``local_optimizer=None``.
        local_n2:
            Minimum number of iterations between consecutive local
            searches. Maximally one local search per performed in each
            iteration. Ignored if ``local_optimizer=None``.
        local_optimizer:
            Local optimizer for refinement, or a callable that creates an
            :class:`pypesto.optimize.Optimizer` or ``None`` to skip local searches.
            In case of a callable, it will be called with the keyword arguments
            `max_walltime_s` and `max_eval`, which should be passed to the optimizer
            (if supported) to honor the overall budget.
            See :class:`SacessFidesFactory` for an example.
        n_diverse:
            Number of samples to choose from to construct the initial RefSet
        max_eval:
            Maximum number of objective functions allowed. This criterion is
            only checked once per iteration, not after every objective
            evaluation, so the actual number of function evaluations may exceed
            this value.
        max_walltime_s:
            Maximum walltime in seconds. Will only be checked between local
            optimizations and other simulations, and thus, may be exceeded by
            the duration of a local search.
        balance:
            Quality vs. diversity balancing factor with
            :math:`0 \leq balance \leq 1`; ``0`` = only quality,
            ``1`` = only diversity.
            Affects the choice of starting points for local searches. I.e.,
            whether local optimization should focus on improving the best
            solutions found so far (quality), or on exploring new regions of
            the parameter space (diversity).
            Ignored if ``local_optimizer=None``.
        n_procs:
            Number of parallel processes to use for parallel function
            evaluation. Mutually exclusive with `n_threads`.
        n_threads:
            Number of parallel threads to use for parallel function evaluation.
            Mutually exclusive with `n_procs`.
        history:
            History of the best values/parameters found so far.
            (Monotonously decreasing objective values.)
        result_includes_refset:
            Whether the :meth:`minimize` result should include the final
            RefSet, or just the local search results and the overall best
            parameters.
        """
        warn(
            "Using pypesto.optimize.ess.ESSOptimizer is deprecated "
            "and will be removed in a future release. "
            "Use pyscat.ESSOptimizer instead "
            "(https://github.com/ICB-DCM/pyscat/, `pip install pyscat`).",
            DeprecationWarning,
            stacklevel=2,
        )

        if max_eval is None and max_walltime_s is None and max_iter is None:
            # in this case, we'd run forever
            raise ValueError(
                "Either `max_iter`, `max_eval` or `max_walltime_s` have to be provided."
            )
        if max_eval is None:
            max_eval = np.inf
        if max_walltime_s is None:
            max_walltime_s = np.inf
        if max_iter is None:
            max_iter = np.inf

        # Hyperparameters
        self.local_n1: int = local_n1
        self.local_n2: int = local_n2
        self.max_iter: int = max_iter
        self.max_eval: int = max_eval
        self.dim_refset: int = dim_refset
        self.local_optimizer = local_optimizer
        self.n_diverse: int = n_diverse
        if n_procs is not None and n_threads is not None:
            raise ValueError(
                "`n_procs` and `n_threads` are mutually exclusive."
            )
        self.n_procs: int | None = n_procs
        self.n_threads: int | None = n_threads
        self.balance: float = balance
        # After how many iterations a stagnated solution is to be replaced by
        #  a random one. Default value taken from [EgeaMar2010]_
        self.n_change: int = 20
        # Only perform local search from best solution
        self.local_only_best_sol: bool = False
        self.max_walltime_s = max_walltime_s
        self._initialize()
        self.logger = logging.getLogger(
            f"{self.__class__.__name__}-{id(self)}"
        )
        self._result_includes_refset = result_includes_refset

    def _initialize(self):
        """(Re-)Initialize."""
        # RefSet
        self.refset: RefSet | None = None
        # Overall best parameters found so far
        self.x_best: np.ndarray | None = None
        # Overall best function value found so far
        self.fx_best: float = np.inf
        # Results from local searches (only those with finite fval)
        # (there is potential to save memory here by only keeping the
        # parameters in memory and not the full result)
        self.local_solutions: list[OptimizerResult] = []
        # Index of current iteration
        self.n_iter: int = 0
        # ESS iteration at which the last local search took place
        # (only local searches with a finite result are counted)
        self.last_local_search_niter: int = 0
        # Whether self.x_best has changed in the current iteration
        self.x_best_has_changed: bool = False
        self.exit_flag: ESSExitFlag = ESSExitFlag.DID_NOT_RUN
        self.evaluator: FunctionEvaluator | None = None
        self.starttime: float | None = None
        self.history: MemoryHistory = MemoryHistory()

    def _initialize_minimize(
        self,
        problem: Problem = None,
        startpoint_method: StartpointMethod = None,
        refset: RefSet | None = None,
    ):
        """Initialize for optimizations.

        Create initial refset, start timer, ... .
        """
        if startpoint_method is not None:
            warn(
                "Passing `startpoint_method` directly is deprecated, "
                "use `problem.startpoint_method` instead.",
                DeprecationWarning,
                stacklevel=1,
            )

        self._initialize()
        self.starttime = time.time()

        if (refset is None and problem is None) or (
            refset is not None and problem is not None
        ):
            raise ValueError(
                "Exactly one of `problem` or `refset` has to be provided."
            )

        problem = problem if problem else refset.evaluator.problem

        if problem.x_guesses.shape[0]:
            # We'll use problem.startpoint_method to sample random points
            #  later on. Depending on the startpoint method, this will return
            #  the provided guesses, meaning that we'll always get the same
            #  points. This means, we won't explore the parameter space, or
            #  potentially even get stuck if the provided guesses are not
            #  evaluable.
            raise ValueError(
                "Providing startpoints in `problem.x_guesses` "
                f"is not supported by {self.__class__.__name__}. "
                "Unset `problem.x_guesses`."
            )

        # generate initial RefSet if not provided
        if refset is None:
            if self.dim_refset is None:
                raise ValueError(
                    "Either refset or dim_refset have to be provided."
                )
            # [EgeaMar2010]_ 2.1
            self.n_diverse = self.n_diverse or 10 * problem.dim
            self.evaluator = create_function_evaluator(
                problem,
                startpoint_method,
                n_threads=self.n_threads,
                n_procs=self.n_procs,
            )

            self.refset = RefSet(dim=self.dim_refset, evaluator=self.evaluator)
            # Initial RefSet generation
            self.refset.initialize_random(n_diverse=self.n_diverse)
        else:
            self.refset = refset

        self.evaluator = self.refset.evaluator
        self.x_best = np.full(
            shape=(self.evaluator.problem.dim,), fill_value=np.nan
        )
        # initialize global best from initial refset
        for x, fx in zip(self.refset.x, self.refset.fx, strict=True):
            self._maybe_update_global_best(x, fx)

    def minimize(
        self,
        problem: Problem = None,
        startpoint_method: StartpointMethod = None,
        refset: RefSet | None = None,
    ) -> pypesto.Result:
        """Minimize the given objective.

        Parameters
        ----------
        problem:
            Problem to run ESS on.
        startpoint_method:
            Method for choosing starting points.
            **Deprecated. Use ``problem.startpoint_method`` instead.**
        refset:
            The initial RefSet or ``None`` to auto-generate.
        """
        self._initialize_minimize(
            problem=problem, startpoint_method=startpoint_method, refset=refset
        )

        # [PenasGon2017]_ Algorithm 1
        while self._keep_going():
            self._do_iteration()

        self._report_final()
        self.history.finalize(exitflag=self.exit_flag.name)
        return self._create_result()

    def _do_iteration(self):
        """Perform an ESS iteration."""
        self.x_best_has_changed = False

        self.refset.sort()
        self._report_iteration()
        self.refset.prune_too_close()

        # Apply combination method to update the RefSet
        x_best_children, fx_best_children = self._combine_solutions()

        # Go-beyond strategy to further improve the new combinations
        self._go_beyond(x_best_children, fx_best_children)

        # Maybe perform a local search
        if self.local_optimizer is not None and self._keep_going():
            self._do_local_search(x_best_children, fx_best_children)

        # Replace RefSet members by best children where an improvement
        #  was made. replace stuck members by random points.
        for i in range(self.refset.dim):
            if fx_best_children[i] < self.refset.fx[i]:
                self.refset.update(i, x_best_children[i], fx_best_children[i])
            else:
                self.refset.n_stuck[i] += 1
                if self.refset.n_stuck[i] > self.n_change:
                    self.refset.replace_by_random(i)

        self.n_iter += 1

    def _create_result(self) -> pypesto.Result:
        """Create the result object.

        Currently, this returns the overall best value and the final RefSet.
        """
        common_result_fields = {
            "exitflag": self.exit_flag,
            # meaningful? this is the overall time, and identical for all
            #  reported points
            "time": time.time() - self.starttime,
            "n_fval": self.evaluator.n_eval,
            "optimizer": str(self),
        }
        i_result = 0
        result = pypesto.Result(problem=self.evaluator.problem)

        # save global best
        optimizer_result = pypesto.OptimizerResult(
            id=str(i_result),
            x=self.x_best,
            fval=self.fx_best,
            message="Global best",
            **common_result_fields,
        )
        optimizer_result.update_to_full(result.problem)
        result.optimize_result.append(optimizer_result)

        # save local solutions
        for i, optimizer_result in enumerate(self.local_solutions):
            i_result += 1
            optimizer_result.id = f"Local solution {i}"
            result.optimize_result.append(optimizer_result)

        if self._result_includes_refset:
            # save refset
            for i in range(self.refset.dim):
                i_result += 1
                result.optimize_result.append(
                    pypesto.OptimizerResult(
                        id=str(i_result),
                        x=self.refset.x[i],
                        fval=self.refset.fx[i],
                        message=f"RefSet[{i}]",
                        **common_result_fields,
                    )
                )
                result.optimize_result[-1].update_to_full(result.problem)

        return result

    def _keep_going(self) -> bool:
        """Check exit criteria.

        Returns
        -------
        ``True`` if not of the exit criteria is met, ``False`` otherwise.
        """
        # TODO DW which further stopping criteria: gtol, fatol, frtol?

        if self.n_iter >= self.max_iter:
            self.exit_flag = ESSExitFlag.MAX_ITER
            return False

        if self._get_remaining_eval() <= 0:
            self.exit_flag = ESSExitFlag.MAX_EVAL
            return False

        if self._get_remaining_time() <= 0:
            self.exit_flag = ESSExitFlag.MAX_TIME
            return False

        return True

    def _get_remaining_time(self):
        """Get remaining wall time in seconds."""
        if self.max_walltime_s is None:
            return np.inf
        return self.max_walltime_s - (time.time() - self.starttime)

    def _get_remaining_eval(self):
        """Get remaining function evaluations."""
        if self.max_eval is None:
            return np.inf
        return self.max_eval - self.evaluator.n_eval

    def _combine_solutions(self) -> tuple[np.ndarray, np.ndarray]:
        """Combine solutions and evaluate.

        Creates the next generation from the RefSet by pair-wise combination
        of all RefSet members. Creates ``RefSet.dim ** 2 - RefSet.dim`` new
        parameter vectors, tests them, and keeps the best child of each parent.

        Returns
        -------
        y:
            The next generation of parameter vectors
            (`dim_refset` x `dim_problem`).
        fy:
            The objective values corresponding to the parameters in `y`.
        """
        y = np.zeros(shape=(self.refset.dim, self.evaluator.problem.dim))
        fy = np.full(shape=self.refset.dim, fill_value=np.inf)
        for i in range(self.refset.dim):
            xs_new = np.vstack(
                tuple(
                    self._combine(i, j)
                    for j in range(self.refset.dim)
                    if i != j
                ),
            )
            fxs_new = self.evaluator.multiple(xs_new)
            best_idx = np.argmin(fxs_new)
            fy[i] = fxs_new[best_idx]
            y[i] = xs_new[best_idx]

            if not self._keep_going():
                break
        return y, fy

    def _combine(self, i, j) -> np.ndarray:
        """Combine RefSet members ``i`` and ``j``.

        Samples a new point from a biased hyper-rectangle derived from the
        given parents, favoring the direction of the better parent.

        Assumes that the RefSet is sorted by quality.

        See [EgeaBal2009]_ Section 3.2 for details.

        Parameters
        ----------
        i:
            Index of first RefSet member for recombination
        j:
            Index of second RefSet member for recombination

        Returns
        -------
        A new parameter vector.
        """
        if i == j:
            raise ValueError("i == j")
        x = self.refset.x

        d = (x[j] - x[i]) / 2
        # i < j implies f(x_i) < f(x_j)
        alpha = 1 if i < j else -1
        # beta is a relative rank-based distance between the two parents
        #  0 <= beta <= 1
        beta = (np.abs(j - i) - 1) / (self.refset.dim - 2)
        # new hyper-rectangle, biased towards the better parent
        c1 = x[i] - d * (1 + alpha * beta)
        c2 = x[i] + d * (1 - alpha * beta)

        # this will not always yield admissible points -> clip to bounds
        ub, lb = self.evaluator.problem.ub, self.evaluator.problem.lb
        c1 = np.fmax(np.fmin(c1, ub), lb)
        c2 = np.fmax(np.fmin(c2, ub), lb)

        return np.random.uniform(
            low=c1, high=c2, size=self.evaluator.problem.dim
        )

    def _do_local_search(
        self, x_best_children: np.ndarray, fx_best_children: np.ndarray
    ) -> None:
        """
        Perform local searches to refine the next generation.

        See [PenasGon2017]_ Algorithm 2.
        """
        if self.local_only_best_sol and self.x_best_has_changed:
            self.logger.debug("Local search only from best point.")
            local_search_x0_fx0_candidates = ((self.x_best, self.fx_best),)
        # first local search?
        elif self.n_iter == self.local_n1:
            self.logger.debug(
                "First local search from best point due to "
                f"local_n1={self.local_n1}."
            )
            local_search_x0_fx0_candidates = ((self.x_best, self.fx_best),)
        elif (
            self.n_iter >= self.local_n1
            and self.n_iter - self.last_local_search_niter >= self.local_n2
        ):
            quality_order = np.argsort(fx_best_children).argsort()
            # compute minimal distance between the best children and all local
            #  optima found so far
            min_distances = (
                np.fromiter(
                    (
                        min(
                            np.linalg.norm(
                                y_i
                                - optimizer_result.x[
                                    optimizer_result.free_indices
                                ]
                            )
                            for optimizer_result in self.local_solutions
                        )
                        for y_i in x_best_children
                    ),
                    dtype=np.float64,
                    count=len(x_best_children),
                )
                if len(self.local_solutions)
                else np.zeros(len(x_best_children))
            )
            # sort by furthest distance to existing local optima
            diversity_order = np.argsort(min_distances)[::-1].argsort()
            # compute priority, balancing quality and diversity
            #  (smaller value = higher priority)
            priority = (
                1 - self.balance
            ) * quality_order + self.balance * diversity_order
            local_search_x0_fx0_candidates = (
                (x_best_children[i], fx_best_children[i])
                for i in np.argsort(priority)
            )
        else:
            return

        # actual local search
        # repeat until a finite value is found, or we don't have any startpoints left
        for (
            local_search_x0,
            local_search_fx0,
        ) in local_search_x0_fx0_candidates:
            optimizer_result = self._local_minimize(
                x0=local_search_x0, fx0=local_search_fx0
            )
            if np.isfinite(optimizer_result.fval):
                self.local_solutions.append(optimizer_result)

                self._maybe_update_global_best(
                    optimizer_result.x[optimizer_result.free_indices],
                    optimizer_result.fval,
                )
                break
        else:
            self.logger.debug(
                "Local search: No finite value found in any local search."
            )
            return

        self.last_local_search_niter = self.n_iter
        self.evaluator.reset_round_counter()

    def _local_minimize(self, x0: np.ndarray, fx0: float) -> OptimizerResult:
        """Perform a local search from the given startpoint."""
        max_walltime_s = self._get_remaining_time()
        max_eval = self._get_remaining_eval()
        # If we are out of budget, return a dummy result.
        # This prevents issues with optimizers that fail if there is no budget
        # (E.g., Ipopt).
        if max_walltime_s < 1 or max_eval < 1:
            msg = "No time or function evaluations left for local search."
            self.logger.info(msg)
            return OptimizerResult(
                id="0",
                x=x0,
                fval=np.inf,
                message=msg,
                n_fval=0,
                n_grad=0,
                time=0,
                history=None,
            )

        # create optimizer instance if necessary
        optimizer = (
            self.local_optimizer
            if isinstance(self.local_optimizer, pypesto.optimize.Optimizer)
            else self.local_optimizer(
                max_eval=max_eval,
                max_walltime_s=max_walltime_s,
            )
        )
        # actual local search
        optimizer_result: OptimizerResult = optimizer.minimize(
            problem=self.evaluator.problem,
            x0=x0,
            id="0",
        )

        # add function evaluations during the local search to our function
        #  evaluation counter (NOTE: depending on the setup, we might neglect
        #  gradient evaluations).
        self.evaluator.n_eval += optimizer_result.n_fval
        self.evaluator.n_eval_round += optimizer_result.n_fval

        self.logger.info(
            f"Local search: {fx0} -> {optimizer_result.fval} "
            f"took {optimizer_result.time:.3g}s, finished with "
            f"{optimizer_result.exitflag}: {optimizer_result.message}"
        )
        return optimizer_result

    def _maybe_update_global_best(self, x, fx):
        """Update the global best value if the provided value is better."""
        if fx < self.fx_best:
            self.x_best[:] = x
            self.fx_best = fx
            self.x_best_has_changed = True
            self.history.update(
                self.x_best.copy(),
                (0,),
                pypesto.C.MODE_FUN,
                {pypesto.C.FVAL: self.fx_best},
            )

    def _go_beyond(self, x_best_children, fx_best_children):
        """Apply go-beyond strategy.

        If a child is better than its parent, intensify search in that
        direction until no further improvement is made.

        See [Egea2009]_ algorithm 1 + section 3.4
        """
        for i in range(self.refset.dim):
            if fx_best_children[i] >= self.refset.fx[i]:
                continue

            # offspring is better than parent
            x_parent = self.refset.x[i].copy()
            fx_parent = self.refset.fx[i]
            x_child = x_best_children[i].copy()
            fx_child = fx_best_children[i]
            improvement = 1
            # Multiplier used in determining the hyper-rectangle from which to
            # sample children. Will be increased in case of 2 consecutive
            # improvements.
            # (corresponds to 1/\Lambda in [Egea2009]_ algorithm 1)
            go_beyond_factor = 1
            while fx_child < fx_parent:
                # update best child
                x_best_children[i] = x_child
                fx_best_children[i] = fx_child

                # create new solution, child becomes parent
                # hyper-rectangle for sampling child
                box_lb = x_child - (x_parent - x_child) * go_beyond_factor
                box_ub = x_child
                # clip to bounds
                ub, lb = self.evaluator.problem.ub, self.evaluator.problem.lb
                box_lb = np.fmax(np.fmin(box_lb, ub), lb)
                box_ub = np.fmax(np.fmin(box_ub, ub), lb)
                # sample parameters
                x_new = np.random.uniform(low=box_lb, high=box_ub)
                x_parent = x_child
                fx_parent = fx_child
                x_child = x_new
                fx_child = self.evaluator.single(x_child)

                improvement += 1
                if improvement == 2:
                    go_beyond_factor *= 2
                    improvement = 0

            # update overall best?
            self._maybe_update_global_best(
                x_best_children[i], fx_best_children[i]
            )
            if not self._keep_going():
                break

    def _report_iteration(self):
        """Log the current iteration."""
        if self.n_iter == 0:
            self.logger.info("iter | best | nf | refset         | nlocal")

        with np.printoptions(
            edgeitems=5,
            threshold=8,
            linewidth=100000,
            formatter={"float": lambda x: f"{x:.3g}"},
        ):
            self.logger.info(
                f"{self.n_iter:4} | {self.fx_best:+.2E} | "
                f"{self.evaluator.n_eval} "
                f"| {self.refset.fx} | {len(self.local_solutions)}"
            )

    def _report_final(self):
        """Log scatter search summary."""
        with np.printoptions(
            edgeitems=5,
            threshold=10,
            linewidth=100000,
            formatter={"float": lambda x: f"{x:.3g}"},
        ):
            self.logger.info(
                f"-- Final ESS fval after {self.n_iter} iterations, "
                f"{self.evaluator.n_eval} function evaluations: {self.fx_best}. "
                f"Exit flag: {self.exit_flag.name}. "
                f"Num local solutions: {len(self.local_solutions)}."
            )
            self.logger.debug(f"Final refset: {np.sort(self.refset.fx)} ")

    def supports_maxtime(self) -> bool:
        """Check whether optimizer supports time limits."""
        return True

    def set_maxtime(self, seconds: float) -> None:
        """Set the maximum wall time for optimization."""
        self.max_walltime_s = seconds

    def supports_maxiter(self) -> bool:
        """Check whether optimizer supports iteration limits."""
        return True

    def set_maxiter(self, iterations: int) -> None:
        """
        Set the maximum number of iterations for optimization.

        Parameters
        ----------
        iterations
            Maximum number of iterations.
        """
        self.max_iter = iterations

    def supports_maxeval(self) -> bool:
        """Check whether optimizer supports evaluation limits."""
        return True

    def set_maxeval(self, evaluations: int) -> None:
        """
        Set the maximum number of function evaluations for optimization.

        Parameters
        ----------
        evaluations
            Maximum number of function evaluations.
        """
        self.max_eval = evaluations



"""AdaptiveParallelTemperingSampler class."""

from collections.abc import Sequence

import numpy as np

from ..C import EXPONENTIAL_DECAY
from .parallel_tempering import ParallelTemperingSampler


class AdaptiveParallelTemperingSampler(ParallelTemperingSampler):
    """Parallel tempering sampler with adaptive temperature adaptation.

    Compared to the base class, this sampler adapts the temperatures
    during the sampling process.
    This both simplifies the setup as it avoids manual tuning,
    and improves the performance as the temperatures are adapted to the
    current state of the chains.

    This implementation is based on:

    * Vousden et al. 2016.
      Dynamic temperature selection for parallel tempering in Markov chain
      Monte Carlo simulations
      (https://doi.org/10.1093/mnras/stv2422),

    via a matlab reference implementation
    (https://github.com/ICB-DCM/PESTO/blob/master/private/performPT.m).
    """

    @classmethod
    def default_options(cls) -> dict:
        """Get default options for sampler."""
        options = super().default_options()
        # scaling factor for temperature adaptation
        options["eta"] = 100
        # controls the adaptation degeneration velocity of the temperature
        # adaption.
        options["nu"] = 1e3
        # initial temperature schedule as in Vousden et. al. 2016.
        options["beta_init"] = EXPONENTIAL_DECAY

        return options

    def adjust_betas(self, i_sample: int, swapped: Sequence[bool]):
        """Update temperatures as in Vousden2016."""
        if len(self.betas) == 1:
            return

        # parameters
        nu = self.options["nu"]
        eta = self.options["eta"]
        betas = self.betas

        # booleans to integer array
        swapped = np.array([int(swap) for swap in swapped])

        # update betas
        kappa = nu / (i_sample + 1 + nu) / eta
        ds = kappa * (swapped[:-1] - swapped[1:])
        dtemp = np.diff(1.0 / betas[:-1])
        dtemp = dtemp * np.exp(ds)
        betas[:-1] = 1 / np.cumsum(np.insert(dtemp, obj=0, values=1.0))

        # fill in
        self.betas = betas



import logging
from collections.abc import Iterable
from typing import SupportsFloat, SupportsInt

import numpy as np

from .base import Problem

SupportsFloatIterableOrValue = Iterable[SupportsFloat] | SupportsFloat
SupportsIntIterableOrValue = Iterable[SupportsInt] | SupportsInt

logger = logging.getLogger(__name__)


class HierarchicalProblem(Problem):
    """
    The Hierarchical Problem.

    A hierarchical problem is a problem with a nested structure: One or
    multiple inner problems are nested inside the outer problem. The inner
    problems are optimized for each evaluation of the outer problem. The
    objective's calculator is used to collect the inner problems' objective
    values.

    Parameters
    ----------
    hierarchical:
        A flag indicating the problem is hierarchical.
    inner_x_names:
        Names of the inner optimization parameters. Only relevant if
        hierarchical is True. Contains the names of easily interpretable
        inner parameters only, e.g. noise parameters, scaling factors, offsets.
    inner_lb, inner_ub:
        The lower and upper bounds for the inner optimization parameters.
        Only relevant if hierarchical is True. Contains the bounds of easily
        interpretable inner parameters only, e.g. noise parameters, scaling
        factors, offsets.
    inner_scales:
        The scales for the inner optimization parameters. Only relevant if
        hierarchical is True. Contains the scales of easily interpretable inner
        parameters only, e.g. noise parameters, scaling factors, offsets. Can
        be pypesto.C.{LIN,LOG,LOG10}. Used only for visualization purposes.
    semiquant_observable_ids:
        The ids of semiquantitative observables. Only relevant if hierarchical
        is True. If not None, the optimization result's `spline_knots` will be
        a list of lists of spline knots for each semiquantitative observable in
        the order of these ids.
    """

    def __init__(
        self,
        inner_x_names: Iterable[str] | None = None,
        inner_lb: np.ndarray | list[float] | None = None,
        inner_ub: np.ndarray | list[float] | None = None,
        **problem_kwargs: dict,
    ):
        super().__init__(**problem_kwargs)

        if inner_x_names is None:
            inner_x_names = (
                self.objective.calculator.get_interpretable_inner_par_ids()
            )
        if len(set(inner_x_names)) != len(inner_x_names):
            raise ValueError("Parameter names inner_x_names must be unique")
        self.inner_x_names = inner_x_names

        if inner_lb is None or inner_ub is None:
            (
                default_inner_lb,
                default_inner_ub,
            ) = self.objective.calculator.get_interpretable_inner_par_bounds()
            inner_lb = default_inner_lb if inner_lb is None else inner_lb
            inner_ub = default_inner_ub if inner_ub is None else inner_ub

        if len(inner_lb) != len(inner_ub):
            raise ValueError("Parameter bounds must have same length")
        if len(inner_lb) != len(inner_x_names):
            raise ValueError(
                "Parameter bounds must have same length as parameter names"
            )

        self.inner_lb = np.array(inner_lb)
        self.inner_ub = np.array(inner_ub)

        self.inner_scales = (
            self.objective.calculator.get_interpretable_inner_par_scales()
        )

        self.semiquant_observable_ids = (
            self.objective.calculator.semiquant_observable_ids
        )
        self.relative_observable_ids = (
            self.objective.calculator.relative_observable_ids
        )



from collections.abc import Iterable, Sequence

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import is_color_like

from pypesto.util import delete_nan_inf

from ..C import COLOR
from ..result import Result
from .clust_color import assign_colors, assign_colors_for_list
from .misc import process_result_list, process_start_indices


def optimization_run_properties_one_plot(
    results: Result,
    properties_to_plot: list[str] | None = None,
    size: tuple[float, float] = (18.5, 10.5),
    start_indices: int | Iterable[int] | None = None,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    legends: str | list[str] | None = None,
    plot_type: str = "line",
) -> matplotlib.axes.Axes:
    """
    Plot stats for allproperties specified in properties_to_plot on one plot.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    properties_to_plot:
        Optimization run properties that should be plotted
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    colors:
        List of colors recognized by matplotlib colors (one color per property in properties_to_plot),
        or single color. If not set and one result, clustering is done
        and colors are assigned automatically
    legends:
        Labels, one label per optimization property
    plot_type:
        Specifies plot type. Possible values: 'line' and 'hist'

    Returns
    -------
    ax:
        The plot axes.

    Examples
    --------
    .. code-block:: python

        optimization_run_properties_one_plot(
            result1,
            properties_to_plot=['time'],
            colors=[.5, .9, .9, .3]
        )

        optimization_run_properties_one_plot(
            result1,
            properties_to_plot=['time', 'n_grad'],
            colors=[[.5, .9, .9, .3], [.2, .1, .9, .5]]
        )
    """
    if properties_to_plot is None:
        properties_to_plot = [
            "time",
            "n_fval",
            "n_grad",
            "n_hess",
            "n_res",
            "n_sres",
        ]

    if colors is None:
        colors = assign_colors_for_list(len(properties_to_plot))
    elif is_color_like(colors):
        colors = [colors]

    if len(colors) != len(properties_to_plot):
        raise ValueError(
            "Number of colors should be the same as number "
            "of optimization properties to plot"
        )

    if legends is None:
        legends = properties_to_plot
    elif not isinstance(legends, list):
        legends = [legends]

    if len(legends) != len(properties_to_plot):
        raise ValueError(
            "Number of legends should be the same as number of "
            "optimization properties to plot"
        )

    ax = plt.subplots()[1]
    fig = plt.gcf()
    fig.set_size_inches(*size)

    for idx, prop_name in enumerate(properties_to_plot):
        optimization_run_property_per_multistart(
            results,
            prop_name,
            ax,
            size,
            start_indices,
            colors[idx],
            legends[idx],
            plot_type,
        )

    ax.set_ylabel("property value")
    ax.set_title("Optimization properties per optimization run")
    return ax


def optimization_run_properties_per_multistart(
    results: Result | Sequence[Result],
    properties_to_plot: list[str] | None = None,
    size: tuple[float, float] = (18.5, 10.5),
    start_indices: int | Iterable[int] | None = None,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    legends: str | list[str] | None = None,
    plot_type: str = "line",
) -> dict[str, plt.Subplot]:
    """
    One plot per optimization property in properties_to_plot.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    properties_to_plot:
        Optimization run properties that should be plotted
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    colors:
        List of colors recognized by matplotlib (one color per result in results),
        or single color. If not set and one result, clustering is done
        and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    plot_type:
        Specifies plot type. Possible values: 'line' and 'hist'

    Returns
    -------
    ax:
    The plot axes.

    Examples
    --------
    .. code-block:: python

        optimization_run_properties_per_multistart(
            result1,
            properties_to_plot=['time'],
            colors=[.5, .9, .9, .3]
        )

        optimization_run_properties_per_multistart(
            [result1, result2],
            properties_to_plot=['time'],
            colors=[[.5, .9, .9, .3], [.2, .1, .9, .5]]
        )

        optimization_run_properties_per_multistart(
            result1,
            properties_to_plot=['time', 'n_grad'],
            colors=[.5, .9, .9, .3]
        )

        optimization_run_properties_per_multistart(
            [result1, result2], properties_to_plot=['time', 'n_fval'],
            colors=[[.5, .9, .9, .3], [.2, .1, .9, .5]]
        )
    """
    if properties_to_plot is None:
        properties_to_plot = [
            "time",
            "n_fval",
            "n_grad",
            "n_hess",
            "n_res",
            "n_sres",
        ]

    num_subplot = len(properties_to_plot)
    # compute, how many rows and columns we need for the subplots
    num_row = int(np.round(np.sqrt(num_subplot)))
    num_col = int(np.ceil(num_subplot / num_row))
    fig, axes = plt.subplots(num_row, num_col, squeeze=False)
    fig.set_size_inches(*size)

    for ax in axes.flat[num_subplot:]:
        ax.remove()
    axes = dict(zip(range(num_subplot), axes.flat, strict=True))
    for idx, prop_name in enumerate(properties_to_plot):
        ax = axes[idx]
        optimization_run_property_per_multistart(
            results,
            prop_name,
            ax,
            size,
            start_indices,
            colors,
            legends,
            plot_type,
        )
    return axes


def optimization_run_property_per_multistart(
    results: Result | Sequence[Result],
    opt_run_property: str,
    axes: matplotlib.axes.Axes | None = None,
    size: tuple[float, float] = (18.5, 10.5),
    start_indices: int | Iterable[int] | None = None,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    legends: str | list[str] | None = None,
    plot_type: str = "line",
) -> matplotlib.axes.Axes:
    """
    Plot stats for an optimization run property specified by opt_run_property.

    It is possible to plot a histogram or a line plot. In a line plot,
    on the x-axis are the numbers of the multistarts, where the multistarts are
    ordered with respect to a function value. On the y-axis of the line plot
    the value of the corresponding parameter for each multistart is displayed.

    Parameters
    ----------
    opt_run_property:
        optimization run property to plot.
        One of the 'time', 'n_fval', 'n_grad', 'n_hess', 'n_res', 'n_sres'
    results:
        Optimization result obtained by 'optimize.py' or list of those
    axes:
        Axes object to use
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    colors:
        List of colors recognized by matplotlib (one color per result in results),
        or single color. If not set and one result, clustering is done
        and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    plot_type:
        Specifies plot type. Possible values: 'line', 'hist', 'both'

    Returns
    -------
    axes:
        The plot axes.
    """
    supported_properties = {
        "time": "Wall-clock time (seconds)",
        "n_fval": "Number of function evaluations",
        "n_grad": "Number of gradient evaluations",
        "n_hess": "Number of Hessian evaluations",
        "n_res": "Number of residuals evaluations",
        "n_sres": "Number of residual sensitivity evaluations",
    }

    if opt_run_property not in supported_properties:
        raise ValueError(
            "Wrong value of opt_run_property. Only the following "
            "values are allowed: 'time', 'n_fval', 'n_grad', "
            "'n_hess', 'n_res', 'n_sres'"
        )

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    # axes
    if axes is None:
        ncols = 2 if plot_type == "both" else 1
        fig, axes = plt.subplots(1, ncols)
        fig.set_size_inches(*size)
        fig.suptitle(
            f"{supported_properties[opt_run_property]} per optimizer run"
        )
    else:
        axes.set_title(
            f"{supported_properties[opt_run_property]} per optimizer run"
        )

    # loop over results
    for j, result in enumerate(results):
        if plot_type == "both":
            axes[0] = stats_lowlevel(
                result,
                opt_run_property,
                supported_properties[opt_run_property],
                axes[0],
                start_indices,
                colors[j],
                legends[j],
            )

            axes[1] = stats_lowlevel(
                result,
                opt_run_property,
                supported_properties[opt_run_property],
                axes[1],
                start_indices,
                colors[j],
                legends[j],
                plot_type="hist",
            )
        else:
            axes = stats_lowlevel(
                result,
                opt_run_property,
                supported_properties[opt_run_property],
                axes,
                start_indices,
                colors[j],
                legends[j],
                plot_type,
            )

    if sum(legend is not None for legend in legends) > 0:
        if plot_type == "both":
            for ax in axes:
                ax.legend()
        else:
            axes.legend()

    return axes


def stats_lowlevel(
    result: Result,
    property_name: str,
    axis_label: str,
    ax: matplotlib.axes.Axes,
    start_indices: int | Iterable[int] | None = None,
    color: COLOR | list[COLOR] | np.ndarray | None = "C0",
    legend: str | None = None,
    plot_type: str = "line",
):
    """
    Plot values of the optimization run property across different multistarts.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'
    property_name:
        name of the optimization result property which value should be plotted
    axis_label:
        Label for the y-axis of the line plot or x-axis of the histogram
    ax:
        Axes object to use
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    color:
        List of colors recognized by matplotlib (length equal to the number of multistarts),
        or single color
        If not set, then for the line plot clustering is done and
        colors are assigned automatically
    legend:
        Label describing the result
    plot_type:
        Specifies plot type. Possible values: 'line' and 'hist'

    Returns
    -------
    ax:
        The plot axes.
    """
    fvals = result.optimize_result.fval
    values = [[res[property_name]] for res in result.optimize_result.list]
    values, fvals = delete_nan_inf(fvals, values)

    if start_indices is not None:
        start_indices = process_start_indices(result, start_indices)
        values = values[start_indices]
        fvals = fvals[start_indices]

    n_starts = len(values)

    # assign colors
    colors = assign_colors(vals=fvals, colors=color, balance_alpha=False)

    sorted_indices = sorted(range(n_starts), key=lambda j: fvals[j])
    values = values[sorted_indices]

    if plot_type == "line":
        # plot line
        ax.plot(range(n_starts), values, color=[0.7, 0.7, 0.7, 0.6])

        # plot points
        for i, v in enumerate(values):
            if i == 0:
                tmp_legend = legend
            else:
                tmp_legend = None
            ax.scatter(i, v, color=colors[i], marker="o", label=tmp_legend)
        ax.set_xlabel("Ordered optimizer run")
        ax.set_ylabel(axis_label)
    else:
        ax.hist(values, color=color, bins="auto", label=legend)
        ax.set_xlabel(axis_label)
        ax.set_ylabel("Number of multistarts")

    return ax



"""Miscellaneous methods."""

import logging
from collections.abc import Iterable
from pathlib import Path

import pandas as pd
import petab.v1 as petab
import petab_select.ui
from petab.v1.C import ESTIMATE, NOMINAL_VALUE
from petab_select import Model, ModelHash, parameter_string_to_value
from petab_select.constants import PETAB_PROBLEM

from ..history import Hdf5History
from ..objective import Objective
from ..optimize import Optimizer
from ..optimize.ess import (
    SacessOptimizer,
    get_default_ess_options,
)
from ..petab import PetabImporter
from ..problem import Problem

logger = logging.getLogger(__name__)


def model_to_pypesto_problem(
    model: Model,
    objective: Objective = None,
    x_guesses: Iterable[dict[str, float]] = None,
    hierarchical: bool = False,
) -> Problem:
    """Create a pyPESTO problem from a PEtab Select model.

    Parameters
    ----------
    model:
        The model.
    objective:
        The pyPESTO objective.
    x_guesses:
        Startpoints to be used in the multi-start optimization. For example,
        this could be the maximum likelihood estimate from another model.
        Each dictionary has parameter IDs as keys, and parameter values as
        values.
        Values in `x_guess` for parameters that are not estimated will be
        ignored and replaced with their value from the PEtab Select model, if
        defined, else their nominal value in the PEtab parameters table.
    hierarchical:
        Whether the problem involves hierarchical optimization.

    Returns
    -------
    The pyPESTO select problem.
    """
    petab_problem = petab_select.ui.model_to_petab(model=model)[PETAB_PROBLEM]

    corrected_x_guesses = None
    if x_guesses is not None:
        corrected_x_guesses = correct_x_guesses(
            x_guesses=x_guesses,
            model=model,
            petab_problem=petab_problem,
            hierarchical=hierarchical,
        )

    importer = PetabImporter(
        petab_problem,
        hierarchical=hierarchical,
    )
    if objective is None:
        factory = importer.create_objective_creator()
        amici_model = factory.create_model(
            non_estimated_parameters_as_constants=False,
        )
        objective = factory.create_objective(
            model=amici_model,
        )
    pypesto_problem = importer.create_problem(
        objective=objective,
        x_guesses=corrected_x_guesses,
    )
    return pypesto_problem


def model_to_hierarchical_pypesto_problem(*args, **kwargs) -> Problem:
    """Create a hierarchical pyPESTO problem from a PEtab Select model.

    See :func:`model_to_pypesto_problem`.
    """
    pypesto_problem = model_to_pypesto_problem(
        *args,
        **kwargs,
        hierarchical=True,
    )
    return pypesto_problem


def correct_x_guesses(
    x_guesses: Iterable[dict[str, float]],
    model: Model,
    petab_problem: petab.Problem = None,
    hierarchical: bool = False,
):
    """Fix startpoint guesses passed between models of different sizes.

    Any parameter values in `x_guess` for parameters that are not estimated
    should be corrected by replacing them with
    - their corresponding values in `row` if possible, else
    - their corresponding nominal values in the `petab_problem.parameter_df`.

    Parameters
    ----------
    x_guesses:
        The startpoints to correct.
    model:
        The startpoints will be corrected to match this model.
    petab_problem:
        The model's corresponding PEtab problem. If this is not provided,
        it will be created from the `model`.
    hierarchical:
        Whether hierarchical optimization is used.

    Returns
    -------
    The corrected startpoint guesses.
    """
    # TODO reconsider whether correcting is a good idea (`x_guess` is no longer
    # the latest MLE then). Similar todo exists in
    # `ModelSelectorMethod.new_model_problem`.
    # TODO move to PEtab Select?
    # TODO may be issues, e.g. differences in bounds of parameters between
    #      different model PEtab problems is not accounted for, or if there are
    #      different parameters in the old/new PEtab problem.

    if petab_problem is None:
        petab_problem = petab_select.ui.model_to_petab(model=model)[
            PETAB_PROBLEM
        ]

    corrected_x_guesses = None
    if x_guesses is not None:
        corrected_x_guesses = []
        for x_guess in x_guesses:
            corrected_x_guess = []
            for parameter_id in petab_problem.parameter_df.index:
                if hierarchical:
                    if not pd.isna(
                        petab_problem.parameter_df.loc[
                            parameter_id, "parameterType"
                        ]
                    ):
                        continue

                # Use the `x_guess` value, if the parameter is to be estimated.
                if (
                    petab_problem.parameter_df[ESTIMATE].loc[parameter_id] == 1
                    and parameter_id in x_guess
                ):
                    corrected_value = x_guess[parameter_id]
                # Else use the PEtab Select model parameter value, if defined.
                elif parameter_id in model.parameters:
                    corrected_value = parameter_string_to_value(
                        model.parameters[parameter_id]
                    )
                # Default to the PEtab parameter table nominal value
                else:
                    corrected_value = petab_problem.parameter_df.loc[
                        parameter_id, NOMINAL_VALUE
                    ]
                corrected_x_guess.append(corrected_value)
            corrected_x_guesses.append(corrected_x_guess)
    return corrected_x_guesses


class SacessMinimizeMethod:
    """Create a minimize method for SaCeSS that adapts to each problem.

    When a pyPESTO SaCeSS optimizer is created, it takes the problem
    dimension as input. Hence, an optimizer needs to be constructed for
    each problem. Objects of this class act like a minimize method for model
    selection, but a new problem-specific SaCeSS optimizer will be created
    every time a model is minimized.

    Instance attributes correspond to pyPESTO's SaCeSS optimizer, and are
    documented there. Extra keyword arguments supplied to the constructor
    will be passed on to the constructor of the SaCeSS optimizer, for example,
    `max_walltime_s` can be specified in this way. If specified, `tmpdir` will
    be treated as a parent directory.
    """

    def __init__(
        self,
        num_workers: int,
        local_optimizer: Optimizer = None,
        tmpdir: str | Path | None = None,
        save_history: bool = False,
        **optimizer_kwargs,
    ):
        """Construct a minimize-like object."""
        self.num_workers = num_workers
        self.local_optimizer = local_optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.save_history = save_history

        self.tmpdir = tmpdir
        if self.tmpdir is not None:
            self.tmpdir = Path(self.tmpdir)

        if self.save_history and self.tmpdir is None:
            self.tmpdir = Path.cwd() / "sacess_tmpdir"

    def __call__(
        self, problem: Problem, model_hash: ModelHash, **minimize_options
    ):
        """Create then run a problem-specific sacess optimizer.

        Parameters
        ----------
        problem:
            The pyPESTO problem for the model.
        model_hash:
            The model hash.
        minimize_options:
            Passed to :meth:`SacessOptimizer.minimize`.

        Returns
        -------
        The output from :meth:`SacessOptimizer.minimize`.
        """
        # create optimizer
        ess_init_args = get_default_ess_options(
            num_workers=self.num_workers,
            dim=problem.dim,
        )
        for x in ess_init_args:
            x["local_optimizer"] = self.local_optimizer
        model_tmpdir = None
        if self.tmpdir is not None:
            model_tmpdir = self.tmpdir / str(model_hash)
            model_tmpdir.mkdir(exist_ok=False, parents=True)

        ess = SacessOptimizer(
            ess_init_args=ess_init_args,
            tmpdir=model_tmpdir,
            **self.optimizer_kwargs,
        )

        # optimize
        result = ess.minimize(
            problem=problem,
            **minimize_options,
        )

        if self.save_history:
            history_dir = model_tmpdir / "history"
            history_dir.mkdir(exist_ok=False, parents=True)
            for history_index, history in enumerate(ess.histories):
                Hdf5History.from_history(
                    other=history,
                    file=history_dir / (str(history_index) + ".hdf5"),
                    id_=history_index,
                )
        return result



"""
Constants
=========
Package-wide consistent constant definitions.
"""

from enum import Enum, StrEnum
from typing import Literal

###############################################################################
# ENSEMBLE

PREDICTOR = "predictor"
PREDICTION_ID = "prediction_id"
PREDICTION_RESULTS = "prediction_results"
PREDICTION_ARRAYS = "prediction_arrays"
PREDICTION_SUMMARY = "prediction_summary"

HISTORY = "history"
OPTIMIZE = "optimize"
SAMPLE = "sample"

MEAN = "mean"
MEDIAN = "median"
STANDARD_DEVIATION = "std"
PERCENTILE = "percentile"
SUMMARY = "summary"
WEIGHTED_SIGMA = "weighted_sigma"

X_NAMES = "x_names"
NX = "n_x"
X_VECTOR = "x_vectors"
NVECTORS = "n_vectors"
VECTOR_TAGS = "vector_tags"
ENSEMBLE_TYPE = "ensemble_type"
PREDICTIONS = "predictions"

SIMULTANEOUS = "simultaneous"
POINTWISE = "pointwise"

LOWER_BOUND = "lower_bound"
UPPER_BOUND = "upper_bound"
PREEQUILIBRATION_CONDITION_ID = "preequilibrationConditionId"
SIMULATION_CONDITION_ID = "simulationConditionId"

COLOR_HIT_BOTH_BOUNDS = [0.6, 0.0, 0.0, 0.9]
COLOR_HIT_ONE_BOUND = [0.95, 0.6, 0.0, 0.9]
COLOR_HIT_NO_BOUNDS = [0.0, 0.8, 0.0, 0.9]


class EnsembleType(Enum):
    """Specifies different ensemble types."""

    ensemble = 1
    sample = 2
    unprocessed_chain = 3


###############################################################################
# OBJECTIVE

MODE_FUN = "mode_fun"  # mode for function values
MODE_RES = "mode_res"  # mode for residuals
ModeType = Literal["mode_fun", "mode_res"]  # type for `mode` argument
FVAL = "fval"  # function value
FVAL0 = "fval0"  # function value at start
GRAD = "grad"  # gradient
HESS = "hess"  # Hessian
HESSP = "hessp"  # Hessian vector product
RES = "res"  # residual
SRES = "sres"  # residual sensitivities
RDATAS = "rdatas"  # returned simulated data sets
OBJECTIVE_NEGLOGPOST = "neglogpost"  # objective is negative log-posterior
OBJECTIVE_NEGLOGLIKE = "negloglike"  # objective is negative log-likelihood

TIME = "time"  # time
N_FVAL = "n_fval"  # number of function evaluations
N_GRAD = "n_grad"  # number of gradient evaluations
N_HESS = "n_hess"  # number of Hessian evaluations
N_RES = "n_res"  # number of residual evaluations
N_SRES = "n_sres"  # number of residual sensitivity evaluations
START_TIME = "start_time"  # start time
X = "x"
X0 = "x0"
ID = "id"

AMICI = "amici"
ROADRUNNER = "roadrunner"
PETAB = "petab"


###############################################################################
# HIERARCHICAL SCALING + OFFSET

INNER_PARAMETERS = "inner_parameters"
PARAMETER_TYPE = "parameterType"
RELATIVE = "relative"


class InnerParameterType(StrEnum):
    """Specifies different inner parameter types."""

    OFFSET = "offset"
    SCALING = "scaling"
    SIGMA = "sigma"
    ORDINAL = "ordinal"
    SPLINE = "spline"


DUMMY_INNER_VALUE = {
    InnerParameterType.OFFSET: 0.0,
    InnerParameterType.SCALING: 1.0,
    InnerParameterType.SIGMA: 1.0,
    InnerParameterType.ORDINAL: 0.0,
    InnerParameterType.SPLINE: 0.0,
}

INNER_PARAMETER_BOUNDS = {
    InnerParameterType.OFFSET: {
        LOWER_BOUND: -float("inf"),
        UPPER_BOUND: float("inf"),
    },
    InnerParameterType.SCALING: {
        LOWER_BOUND: -float("inf"),
        UPPER_BOUND: float("inf"),
    },
    InnerParameterType.SIGMA: {
        LOWER_BOUND: 0,
        UPPER_BOUND: float("inf"),
    },
    InnerParameterType.ORDINAL: {
        LOWER_BOUND: -float("inf"),
        UPPER_BOUND: float("inf"),
    },
    InnerParameterType.SPLINE: {
        LOWER_BOUND: -float("inf"),
        UPPER_BOUND: float("inf"),
    },
}

###############################################################################
# OPTIMAL SCALING

# Should go to PEtab constants at some point
MEASUREMENT_CATEGORY = "measurementCategory"
MEASUREMENT_TYPE = "measurementType"
CENSORING_BOUNDS = "censoringBounds"

ORDINAL = "ordinal"
CENSORED = "censored"
LEFT_CENSORED = "left-censored"
RIGHT_CENSORED = "right-censored"
INTERVAL_CENSORED = "interval-censored"
CENSORING_TYPES = [LEFT_CENSORED, RIGHT_CENSORED, INTERVAL_CENSORED]

REDUCED = "reduced"
STANDARD = "standard"
MAXMIN = "max-min"
MAX = "max"

METHOD = "method"
REPARAMETERIZED = "reparameterized"
INTERVAL_CONSTRAINTS = "interval_constraints"
MIN_GAP = "min_gap"
ORDINAL_OPTIONS = [
    METHOD,
    REPARAMETERIZED,
    INTERVAL_CONSTRAINTS,
    MIN_GAP,
]

CAT_LB = "cat_lb"
CAT_UB = "cat_ub"

NUM_CATEGORIES = "num_categories"
NUM_DATAPOINTS = "num_datapoints"
SURROGATE_DATA = "surrogate_data"
NUM_INNER_PARAMS = "num_inner_params"
LB_INDICES = "lb_indices"
UB_INDICES = "ub_indices"

QUANTITATIVE_IXS = "quantitative_ixs"
QUANTITATIVE_DATA = "quantitative_data"
NUM_CONSTR_FULL = "num_constr_full"
C_MATRIX = "C_matrix"
W_MATRIX = "W_matrix"
W_DOT_MATRIX = "W_dot_matrix"

SCIPY_SUCCESS = "success"
SCIPY_FUN = "fun"
SCIPY_X = "x"

###############################################################################
# SPLINE APPROXIMATION FOR SEMIQUANTITATIVE DATA

MEASUREMENT_TYPE = "measurementType"

SEMIQUANTITATIVE = "semiquantitative"

SPLINE_RATIO = "spline_ratio"
MIN_DIFF_FACTOR = "min_diff_factor"
REGULARIZE_SPLINE = "regularize_spline"
REGULARIZATION_FACTOR = "regularization_factor"
SPLINE_APPROXIMATION_OPTIONS = [
    SPLINE_RATIO,
    MIN_DIFF_FACTOR,
    REGULARIZE_SPLINE,
    REGULARIZATION_FACTOR,
]

MIN_SIM_RANGE = 1e-16

SPLINE_PAR_TYPE = "spline"
SPLINE_KNOTS = "spline_knots"
N_SPLINE_PARS = "n_spline_pars"
DATAPOINTS = "datapoints"
MIN_DATAPOINT = "min_datapoint"
MAX_DATAPOINT = "max_datapoint"
EXPDATA_MASK = "expdata_mask"
CURRENT_SIMULATION = "current_simulation"
INNER_NOISE_PARS = "inner_noise_pars"
OPTIMIZE_NOISE = "optimize_noise"


###############################################################################
# HISTORY

HISTORY = "history"
TRACE = "trace"
N_ITERATIONS = "n_iterations"
MESSAGES = "messages"
MESSAGE = "message"
EXITFLAG = "exitflag"
TRACE_SAVE_ITER = "trace_save_iter"

SUFFIXES_CSV = ["csv"]
SUFFIXES_HDF5 = ["hdf5", "h5"]
SUFFIXES = SUFFIXES_CSV + SUFFIXES_HDF5

CPU_TIME_TOTAL = "cpu_time_total"
PREEQ_CPU_TIME = "preeq_cpu_time"
PREEQ_CPU_TIME_BACKWARD = "preeq_cpu_time_b"
POSTEQ_CPU_TIME = "posteq_cpu_time"
POSTEQ_CPU_TIME_BACKWARD = "posteq_cpu_time_b"


###############################################################################
# PRIOR

LIN = "lin"  # linear
LOG = "log"  # logarithmic to basis e
LOG10 = "log10"  # logarithmic to basis 10

UNIFORM = "uniform"
PARAMETER_SCALE_UNIFORM = "parameterScaleUniform"
NORMAL = "normal"
PARAMETER_SCALE_NORMAL = "parameterScaleNormal"
LAPLACE = "laplace"
PARAMETER_SCALE_LAPLACE = "parameterScaleLaplace"
LOG_UNIFORM = "logUniform"
LOG_NORMAL = "logNormal"
LOG_LAPLACE = "logLaplace"

###############################################################################
# SAMPLING

EXPONENTIAL_DECAY = (
    "exponential_decay"  # temperature schedule for parallel tempering
)
BETA_DECAY = "beta_decay"  # temperature schedule for parallel tempering
TRAPEZOID = "trapezoid"  # method to compute log evidence
SIMPSON = "simpson"  # method to compute log evidence
STEPPINGSTONE = "steppingstone"  # method to compute log evidence

###############################################################################
# PREDICT

OUTPUT_IDS = "output_ids"  # data member in PredictionConditionResult
PARAMETER_IDS = "x_names"  # data member in PredictionConditionResult
TIMEPOINTS = "timepoints"  # data member in PredictionConditionResult
OUTPUT = "output"  # field in the return dict of AmiciPredictor
OUTPUT_SENSI = "output_sensi"  # field in the return dict of AmiciPredictor
OUTPUT_WEIGHT = "output_weight"  # field in the return dict of AmiciPredictor
OUTPUT_SIGMAY = "output_sigmay"  # field in the return dict of AmiciPredictor

# separator in the conditions_ids between preequilibration and simulation
# condition
CONDITION_SEP = "::"

AMICI_T = "t"  # return field in amici simulation result
AMICI_X = "x"  # return field in amici simulation result
AMICI_SX = "sx"  # return field in amici simulation result
AMICI_Y = "y"  # return field in amici simulation result
AMICI_SY = "sy"  # return field in amici simulation result
AMICI_LLH = "llh"  # return field in amici simulation result
AMICI_STATUS = "status"  # return field in amici simulation result
AMICI_SIGMAY = "sigmay"  # return field in amici simulation result
AMICI_SSIGMAY = "ssigmay"  # return field in amici simulation result
AMICI_SSIGMAZ = "ssigmaz"  # return field in amici simulation result

ROADRUNNER_LLH = "llh"  # return field in roadrunner objective
ROADRUNNER_INSTANCE = "roadrunner_instance"
ROADRUNNER_SIMULATION = "simulation_results"

CONDITION = "condition"
CONDITION_IDS = "condition_ids"

CSV = "csv"  # return file format
H5 = "h5"  # return file format

###############################################################################
# VISUALIZE

LEN_RGB = 3  # number of elements in an RGB color
LEN_RGBA = 4  # number of elements in an RGBA color
RGB = tuple[(float,) * LEN_RGB]  # typing of an RGB color
RGBA = tuple[(float,) * LEN_RGBA]  # typing of an RGBA color
RGB_RGBA = RGB | RGBA  # typing of an RGB or RGBA color
RGBA_MIN = 0  # min value for an RGBA element
RGBA_MAX = 1  # max value for an RGBA element
RGBA_ALPHA = 3  # zero-indexed fourth element in RGBA
RGBA_WHITE = (RGBA_MAX, RGBA_MAX, RGBA_MAX, RGBA_MAX)  # white as an RGBA color
RGBA_BLACK = (RGBA_MIN, RGBA_MIN, RGBA_MIN, RGBA_MAX)  # black as an RGBA color
COLOR = str | RGB | RGBA  # typing of a color recognized by matplotlib

# optimizer history
TRACE_X_TIME = "time"
TRACE_X_STEPS = "steps"
# supported values to plot on x-axis
TRACE_X = (TRACE_X_TIME, TRACE_X_STEPS)

TRACE_Y_FVAL = "fval"
TRACE_Y_GRADNORM = "gradnorm"
# supported values to plot on y-axis
TRACE_Y = (TRACE_Y_FVAL, TRACE_Y_GRADNORM)

# parameter indices
FREE_ONLY = "free_only"  # only estimated parameters
ALL = "all"  # all parameters, also for start indices

# start indices
ALL_CLUSTERED = "all_clustered"  # best + all that are in a cluster of size > 1
FIRST_CLUSTER = "first_cluster"  # all starts that belong to the first cluster

# waterfall max value
WATERFALL_MAX_VALUE = 1e100


###############################################################################
# ENVIRONMENT VARIABLES

PYPESTO_MAX_N_STARTS: str = "PYPESTO_MAX_N_STARTS"
PYPESTO_MAX_N_SAMPLES: str = "PYPESTO_MAX_N_SAMPLES"



from __future__ import annotations

import copy
from typing import Any

import numpy as np

from ...C import InnerParameterType
from ...objective import Objective
from ...objective.amici.amici_util import add_sim_grad_to_opt_grad
from ...optimize import minimize
from ...problem import Problem
from ..base_parameter import InnerParameter
from ..base_problem import (
    InnerProblem,
    scale_back_value_dict,
    scale_value_dict,
)
from ..base_solver import InnerSolver
from .util import (
    apply_offset,
    apply_scaling,
    apply_scaling_to_sensitivities,
    apply_sigma,
    compute_bounded_optimal_scaling_offset_coupled,
    compute_nllh,
    compute_nllh_gradient_for_condition,
    compute_optimal_offset,
    compute_optimal_offset_coupled,
    compute_optimal_scaling,
    compute_optimal_sigma,
)

try:
    import amici
    from amici.sim._parameter_mapping import ParameterMapping
except ImportError:
    pass


class RelativeInnerSolver(InnerSolver):
    """Base class for the relative inner solver."""

    def calculate_obj_function(
        self,
        problem: InnerProblem,
        sim: list[np.ndarray],
        sigma: list[np.ndarray],
        inner_parameters: dict[str, float],
    ) -> float:
        """Calculate the objective function value.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        """
        relevant_data = copy.deepcopy(problem.data)
        sim = copy.deepcopy(sim)
        sigma = copy.deepcopy(sigma)
        inner_parameters = copy.deepcopy(inner_parameters)
        inner_parameters = scale_back_value_dict(inner_parameters, problem)

        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            apply_offset(
                offset_value=inner_parameters[x.inner_parameter_id],
                data=relevant_data,
                mask=x.ixs,
            )

        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            apply_scaling(
                scaling_value=inner_parameters[x.inner_parameter_id],
                sim=sim,
                mask=x.ixs,
            )

        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            apply_sigma(
                sigma_value=inner_parameters[x.inner_parameter_id],
                sigma=sigma,
                mask=x.ixs,
            )

        return compute_nllh(relevant_data, sim, sigma, problem.data_mask)

    def calculate_gradients(
        self,
        problem: InnerProblem,
        sim: list[np.ndarray],
        ssim: list[np.ndarray],
        sigma: list[np.ndarray],
        ssigma: list[np.ndarray],
        inner_parameters: dict[str, float],
        parameter_mapping: ParameterMapping,
        par_opt_ids: list[str],
        par_sim_ids: list[str],
        snllh: np.ndarray,
    ) -> np.ndarray:
        """Calculate the gradients with respect to the outer parameters.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        ssim:
            List of sensitivity matrices from the model, as provided in AMICI's
            ``ReturnData.sy``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        ssigma:
            List of sigma sensitivity matrices from the model, as provided
            in AMICI's ``ReturnData.ssigmay``. Same order as simulations in the
            PEtab problem.
        inner_parameters:
            The computed inner parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        par_opt_ids:
            Ids of outer otimization parameters.
        par_sim_ids:
            Ids of outer simulation parameters, includes fixed parameters.
        snllh:
            A zero-initialized vector of the same length as ``par_opt_ids`` to store the
            gradients in. Will be modified in-place.

        Returns
        -------
        The gradients with respect to the outer parameters.
        """
        relevant_data = copy.deepcopy(problem.data)
        sim = copy.deepcopy(sim)
        sigma = copy.deepcopy(sigma)
        inner_parameters = copy.deepcopy(inner_parameters)
        inner_parameters = scale_back_value_dict(inner_parameters, problem)

        # restructure sensitivities to have parameter index as second index
        ssim = [
            np.asarray(
                [
                    ssim[cond_idx][:, par_idx, :]
                    for par_idx in range(ssim[cond_idx].shape[1])
                ]
            )
            for cond_idx in range(len(sim))
        ]
        ssigma = [
            np.asarray(
                [
                    ssigma[cond_idx][:, par_idx, :]
                    for par_idx in range(ssigma[cond_idx].shape[1])
                ]
            )
            for cond_idx in range(len(sigma))
        ]

        # apply offsets, scalings and sigmas
        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            apply_offset(
                offset_value=inner_parameters[x.inner_parameter_id],
                data=relevant_data,
                mask=x.ixs,
            )

        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            apply_scaling(
                scaling_value=inner_parameters[x.inner_parameter_id],
                sim=sim,
                mask=x.ixs,
            )
            apply_scaling_to_sensitivities(
                scaling_value=inner_parameters[x.inner_parameter_id],
                ssim=ssim,
                mask=x.ixs,
            )

        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            apply_sigma(
                sigma_value=inner_parameters[x.inner_parameter_id],
                sigma=sigma,
                mask=x.ixs,
            )

        # compute gradients
        for cond_idx, cond_par_map in enumerate(parameter_mapping):
            gradient_for_cond = compute_nllh_gradient_for_condition(
                data=relevant_data[cond_idx],
                sim=sim[cond_idx],
                ssim=ssim[cond_idx],
                sigma=sigma[cond_idx],
                ssigma=ssigma[cond_idx],
            )
            add_sim_grad_to_opt_grad(
                par_opt_ids=par_opt_ids,
                par_sim_ids=par_sim_ids,
                condition_map_sim_var=cond_par_map.map_sim_var,
                sim_grad=gradient_for_cond,
                opt_grad=snllh,
            )

        return snllh

    def apply_inner_parameters_to_rdatas(
        self,
        problem: InnerProblem,
        rdatas: list[amici.ReturnData],
        inner_parameters: dict[str, float],
    ):
        """Apply the inner parameters to the rdatas.

        If we have simulated the model with dummy inner parameters only,
        we need to apply the inner parameters to the rdatas to get the
        correct model output.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        rdatas:
            The rdatas to apply the inner parameters to.
        inner_parameters:
            The inner parameters to apply to the rdatas.
        """
        sim = [rdata["y"] for rdata in rdatas]
        sigma = [rdata["sigmay"] for rdata in rdatas]
        inner_parameters = copy.deepcopy(inner_parameters)
        inner_parameters = scale_back_value_dict(inner_parameters, problem)

        # apply offsets, scalings and sigmas
        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            apply_scaling(
                scaling_value=inner_parameters[x.inner_parameter_id],
                sim=sim,
                mask=x.ixs,
            )

        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            apply_offset(
                offset_value=inner_parameters[x.inner_parameter_id],
                data=sim,
                mask=x.ixs,
                is_data=False,
            )

        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            apply_sigma(
                sigma_value=inner_parameters[x.inner_parameter_id],
                sigma=sigma,
                mask=x.ixs,
            )

        return rdatas


class AnalyticalInnerSolver(RelativeInnerSolver):
    """Solve the inner subproblem analytically.

    Currently, supports sigmas for additive Gaussian noise.
    """

    def solve(
        self,
        problem: InnerProblem,
        sim: list[np.ndarray],
        sigma: list[np.ndarray],
        scaled: bool,
    ) -> dict[str, float]:
        """Solve the subproblem analytically.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        scaled:
            Whether to scale the results to the parameter scale specified in
            ``problem``.
        """
        x_opt = {}
        data = copy.deepcopy(problem.data)
        sim = copy.deepcopy(sim)
        sigma = copy.deepcopy(sigma)

        # compute optimal offsets
        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            if x.coupled is not None:
                x_opt[x.inner_parameter_id] = compute_optimal_offset_coupled(
                    data=data, sim=sim, sigma=sigma, mask=x.ixs
                )

                # calculate the optimal coupled scaling
                coupled_scaling = x.coupled
                x_opt[coupled_scaling.inner_parameter_id] = (
                    compute_optimal_scaling(
                        data=data,
                        sim=sim,
                        sigma=sigma,
                        mask=coupled_scaling.ixs,
                        optimal_offset=x_opt[x.inner_parameter_id],
                    )
                )

                # check whether they both satisfy their bounds
                if x.is_within_bounds(
                    x_opt[x.inner_parameter_id]
                ) and coupled_scaling.is_within_bounds(
                    x_opt[coupled_scaling.inner_parameter_id]
                ):
                    continue
                else:
                    # if not, we need to recompute them
                    (
                        x_opt[coupled_scaling.inner_parameter_id],
                        x_opt[x.inner_parameter_id],
                    ) = compute_bounded_optimal_scaling_offset_coupled(
                        data=data,
                        sim=sim,
                        sigma=sigma,
                        s=coupled_scaling,
                        b=x,
                        s_opt_value=x_opt[coupled_scaling.inner_parameter_id],
                        b_opt_value=x_opt[x.inner_parameter_id],
                    )
            # compute non-coupled optimal offset
            else:
                x_opt[x.inner_parameter_id] = compute_optimal_offset(
                    data=data, sim=sim, sigma=sigma, mask=x.ixs
                )
                # check if the solution is within bounds
                # if not, we set it to the unsatisfied bound
                if not x.is_within_bounds(x_opt[x.inner_parameter_id]):
                    x_opt[x.inner_parameter_id] = x.get_bounds()[
                        x.get_unsatisfied_bound(x_opt[x.inner_parameter_id])
                    ]

        # apply offsets
        for x in problem.get_xs_for_type(InnerParameterType.OFFSET):
            apply_offset(
                offset_value=x_opt[x.inner_parameter_id], data=data, mask=x.ixs
            )

        # compute non-coupled optimal scalings
        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            if x.coupled is None:
                x_opt[x.inner_parameter_id] = compute_optimal_scaling(
                    data=data, sim=sim, sigma=sigma, mask=x.ixs
                )
                # check if the solution is within bounds
                # if not, we set it to the unsatisfied bound
                if not x.is_within_bounds(x_opt[x.inner_parameter_id]):
                    x_opt[x.inner_parameter_id] = x.get_bounds()[
                        x.get_unsatisfied_bound(x_opt[x.inner_parameter_id])
                    ]
        # apply scalings
        for x in problem.get_xs_for_type(InnerParameterType.SCALING):
            apply_scaling(
                scaling_value=x_opt[x.inner_parameter_id], sim=sim, mask=x.ixs
            )

        # compute optimal sigmas
        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            x_opt[x.inner_parameter_id] = compute_optimal_sigma(
                data=data, sim=sim, mask=x.ixs
            )
        # apply sigmas
        for x in problem.get_xs_for_type(InnerParameterType.SIGMA):
            apply_sigma(
                sigma_value=x_opt[x.inner_parameter_id],
                sigma=sigma,
                mask=x.ixs,
            )

        # scale
        if scaled:
            x_opt = scale_value_dict(x_opt, problem)
        return x_opt


class NumericalInnerSolver(RelativeInnerSolver):
    """Solve the inner subproblem numerically.

    Advantage: The structure of the subproblem does not matter like, at all.
    Disadvantage: Slower.

    Special features: We cache the best parameters, which substantially
    speeds things up.

    Attributes
    ----------
    minimize_kwargs:
        Passed to the `pypesto.optimize.minimize` call.
    n_cached:
        Number of optimized parameter vectors to save.
    problem_kwargs:
        Passed to the `pypesto.Problem` constructor.
    x_guesses:
        Cached optimized parameter vectors, supplied as guesses to the next
        `solve` call.
    """

    def __init__(
        self,
        minimize_kwargs: dict[str, Any] = None,
        n_cached: int = 1,
        problem_kwargs: dict[str, Any] = None,
    ):
        self.minimize_kwargs = minimize_kwargs
        if self.minimize_kwargs is None:
            self.minimize_kwargs = {}
        self.n_cached = n_cached
        self.problem_kwargs = problem_kwargs
        if self.problem_kwargs is None:
            self.problem_kwargs = {}

        self.minimize_kwargs["n_starts"] = self.minimize_kwargs.get(
            "n_starts", 1
        )
        self.minimize_kwargs["progress_bar"] = self.minimize_kwargs.get(
            "progress_bar", False
        )

        self.x_guesses = None
        self.dummy_lb = -1e20
        self.dummy_ub = +1e20
        self.user_specified_lb = None
        self.user_specified_ub = None

    def initialize(self):
        """(Re-)initialize the solver."""
        self.x_guesses = None

    def solve(
        self,
        problem: InnerProblem,
        sim: list[np.ndarray],
        sigma: list[np.ndarray],
        scaled: bool,
    ) -> dict[str, float]:
        """Solve the subproblem numerically.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        sim:
            List of model output matrices, as provided in AMICI's
            ``ReturnData.y``. Same order as simulations in the
            PEtab problem.
        sigma:
            List of sigma matrices from the model, as provided in AMICI's
            ``ReturnData.sigmay``. Same order as simulations in the
            PEtab problem.
        scale:
            Whether to scale the results to the parameter scale specified in
            ``problem``.
        """
        pars = list(problem.xs.values())

        # This has to be done only once
        if self.user_specified_lb is None or self.user_specified_ub is None:
            self.user_specified_lb = [
                i for i in range(len(pars)) if pars[i].lb != -np.inf
            ]
            self.user_specified_ub = [
                i for i in range(len(pars)) if pars[i].ub != np.inf
            ]

        lb = [x.lb for x in pars]
        ub = [x.ub for x in pars]

        x_guesses = self.sample_startpoints(problem, pars)

        x_names = [x.inner_parameter_id for x in pars]
        data = problem.data

        # objective function
        def fun(x):
            _sim = copy.deepcopy(sim)
            _sigma = copy.deepcopy(sigma)
            _data = copy.deepcopy(data)
            for x_val, par in zip(x, pars, strict=True):
                mask = par.ixs
                if par.inner_parameter_type == InnerParameterType.OFFSET:
                    apply_offset(x_val, _data, mask)
                elif par.inner_parameter_type == InnerParameterType.SCALING:
                    apply_scaling(x_val, _sim, mask)
                elif par.inner_parameter_type == InnerParameterType.SIGMA:
                    apply_sigma(x_val, _sigma, mask)
                else:
                    raise ValueError(
                        "Can't handle parameter type "
                        f"`{par.inner_parameter_type}`."
                    )

            return compute_nllh(_data, _sim, _sigma, problem.data_mask)

        # TODO gradient
        objective = Objective(fun)

        # optimization problem
        pypesto_problem = Problem(
            objective, lb=lb, ub=ub, x_names=x_names, **self.problem_kwargs
        )
        pypesto_problem.set_x_guesses(
            x_guesses[:, pypesto_problem.x_free_indices]
        )

        # perform the actual optimization
        result = minimize(pypesto_problem, **self.minimize_kwargs)
        best_par = result.optimize_result.list[0]["x"]

        # Check if the index of an optimized parameter on the dummy bound
        # is not in the list of specified bounds. If so, raise an error.
        if any(
            (
                i not in self.user_specified_lb
                for i, x in enumerate(best_par)
                if x == self.dummy_lb
            )
        ) or any(
            (
                i not in self.user_specified_ub
                for i, x in enumerate(best_par)
                if x == self.dummy_ub
            )
        ):
            raise RuntimeError(
                f"An optimal inner parameter is on the default dummy bound of numerical optimization. "
                f"This means the optimal inner parameter is either extremely large (>={self.dummy_ub})"
                f"or extremely small (<={self.dummy_lb}). Consider changing the inner parameter bounds."
            )

        x_opt = dict(zip(pypesto_problem.x_names, best_par, strict=True))

        # cache
        self.x_guesses = np.array(
            [
                entry["x"]
                for entry in result.optimize_result.list[: self.n_cached]
            ]
        )

        # scale
        if scaled:
            x_opt = scale_value_dict(x_opt, problem)

        return x_opt

    def sample_startpoints(
        self, problem: InnerProblem, pars: list[InnerParameter]
    ) -> np.ndarray:
        """Sample startpoints for the numerical optimization.

        Samples the startpoints for the numerical optimization from a
        log-uniform distribution using the symmetric logarithmic scale.

        Parameters
        ----------
        problem:
            The inner problem to solve.
        pars:
            The inner parameters to sample startpoints for.

        Returns
        -------
        The sampled startpoints appended to the cached startpoints.
        """
        if self.minimize_kwargs["n_starts"] == 1 and self.x_guesses is None:
            return np.array(
                [list(problem.get_dummy_values(scaled=False).values())]
            )
        elif self.x_guesses is not None:
            n_samples = self.minimize_kwargs["n_starts"] - len(self.x_guesses)
        else:
            n_samples = self.minimize_kwargs["n_starts"] - 1

        if n_samples <= 0:
            return self.x_guesses

        lb = np.nan_to_num([x.lb for x in pars], neginf=self.dummy_lb)
        ub = np.nan_to_num([x.ub for x in pars], posinf=self.dummy_ub)

        def symlog10(x):
            return np.sign(x) * np.log10(np.abs(x) + 1)

        def inverse_symlog10(x):
            return np.sign(x) * (np.power(10, np.abs(x)) - 1)

        # Sample startpoints from a log-uniform distribution
        startpoints = np.random.uniform(
            low=symlog10(lb),
            high=symlog10(ub),
            size=(n_samples, len(pars)),
        )
        startpoints = inverse_symlog10(startpoints)

        # Stack the sampled startpoints with the cached startpoints
        if self.x_guesses is not None:
            startpoints = np.vstack(
                (
                    self.x_guesses,
                    startpoints,
                )
            )
        else:
            startpoints = np.vstack(
                (
                    np.array(
                        [list(problem.get_dummy_values(scaled=False).values())]
                    ),
                    startpoints,
                )
            )
        return startpoints



import copy
import warnings

import numpy as np

from ...C import (
    DUMMY_INNER_VALUE,
    LOWER_BOUND,
    UPPER_BOUND,
    InnerParameterType,
)
from ..base_parameter import InnerParameter


def get_finite_quotient(
    numerator: float,
    denominator: float,
    inner_parameter_type: InnerParameterType,
):
    """Get a finite value for the inner parameter.

    Parameters
    ----------
    numerator:
        The numerator of a quotient.
    denominator:
        The denominator of a quotient.
    useless:
        The value to return if the quotient is not finite.

    Returns
    -------
    `num / den` if it's finite, else a dummy value.
    """
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            quotient = float(numerator / denominator)
        if not np.isfinite(quotient):
            raise ValueError
        return quotient
    except Exception:
        return DUMMY_INNER_VALUE[inner_parameter_type]


def compute_optimal_scaling(
    data: list[np.ndarray],
    sim: list[np.ndarray],
    sigma: list[np.ndarray],
    mask: list[np.ndarray],
    optimal_offset: float = None,
) -> float:
    """
    Compute optimal scaling.

    Compute optimal scaling parameter for the given measurements and model
    outputs. See https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.1
    for the derivation.
    """
    # SI Section 3.1: page 9, bottom (s = ...) with offset b=0 (as there is
    # no offset, or it already has been subtracted from the measurements)
    # numerator and denominator to compute the optimal scaling
    # num: \sum_i \frac{ \bar{y}_i * \tilde{h}_i } { \sigma_i^2 }
    # den:  \sum_i \frac{ \tilde{h}_i^2 }{ \sigma_i^2 }

    # numerator, denominator
    num, den = 0.0, 0.0

    # iterate over conditions
    for sim_i, data_i, sigma_i, mask_i in zip(
        sim, data, sigma, mask, strict=True
    ):
        # extract relevant values
        sim_x = sim_i[mask_i]  # \tilde{h}_i
        data_x = (
            data_i[mask_i] - optimal_offset
            if optimal_offset is not None
            else data_i[mask_i]
        )  # \bar{y}_i
        sigma_x = sigma_i[mask_i]  # \sigma_i
        # update statistics
        num += np.nansum(sim_x * data_x / sigma_x**2)
        den += np.nansum(sim_x**2 / sigma_x**2)

    return get_finite_quotient(
        numerator=num,
        denominator=den,
        inner_parameter_type=InnerParameterType.SCALING,
    )


def apply_scaling(
    scaling_value: float, sim: list[np.ndarray], mask: list[np.ndarray]
):
    """Apply scaling to simulations (in-place).

    Parameters
    ----------
    scaling_value:
        The optimal scaling for the masked simulations.
    sim:
        All full (unmasked) simulations.
    mask:
        The masks that indicate the simulation subset that corresponds to the
        `scaling_value`.
    """
    for i in range(len(sim)):
        sim[i][mask[i]] *= scaling_value


def apply_scaling_to_sensitivities(
    scaling_value: float, ssim: list[np.ndarray], mask: list[np.ndarray]
):
    """Apply scaling to sensitivities (in-place).

    Parameters
    ----------
    scaling_value:
        The optimal scaling for the masked simulations.
    ssim:
        All full (unmasked) sensitivities.
    mask:
        The masks that indicate the simulation subset that corresponds to the
        `scaling_value`.
    """
    for i in range(len(ssim)):
        ssim[i][:, mask[i]] *= scaling_value


def compute_optimal_offset(
    data: list[np.ndarray],
    sim: list[np.ndarray],
    sigma: list[np.ndarray],
    mask: list[np.ndarray],
    optimal_scaling: float = None,
) -> float:
    """Compute optimal offset.

    Compute optimal offset for the given measurements and model outputs. See
    https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.1 for the
    derivation. This function handles offsets that occur without any coupled
    scaling parameter.
    """
    # SI Section 3.1: page 9, bottom (b = ...) with scaling s=1
    # numerator and denominator to compute the optimal offset
    # num: \sum_i \frac{ \bar{y}_i - \tilde{h}_i } { \sigma_i^2 }
    # den:  \sum_i \frac{1}{ \sigma_i^2 }
    num, den = 0.0, 0.0

    # iterate over conditions
    for sim_i, data_i, sigma_i, mask_i in zip(
        sim, data, sigma, mask, strict=True
    ):
        # extract relevant values
        sim_x = (
            optimal_scaling * sim_i[mask_i]
            if optimal_scaling is not None
            else sim_i[mask_i]
        )  # \tilde{h}_i
        data_x = data_i[mask_i]  # \bar{y}_i
        sigma_x = sigma_i[mask_i]  # \sigma_i
        # update statistics
        num += np.nansum((data_x - sim_x) / sigma_x**2)
        den += np.nansum(1 / sigma_x**2)
    return get_finite_quotient(
        numerator=num,
        denominator=den,
        inner_parameter_type=InnerParameterType.OFFSET,
    )


def compute_optimal_offset_coupled(
    data: list[np.ndarray],
    sim: list[np.ndarray],
    sigma: list[np.ndarray],
    mask: list[np.ndarray],
) -> float:
    """Compute optimal offset.

    Compute optimal offset for an observable that has both an offset and
    scaling inner parameter.

    See https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.1 for the
    derivation.
    """
    # will be \sum_i \frac{ \tilde{h}_i }{ \sigma_i^2 }
    h = 0.0
    # will be \sum_i \frac{1}{ \sigma_i^2 }
    recnoise = 0.0
    # will be \sum_i \frac{ \bar{y}_i * \tilde{h}_i }{ \sigma_i^2 }
    yh = 0.0
    # will be \sum_i \frac{ \bar{y}_i }{ \sigma_i^2 }
    y = 0.0
    # will be \sum_i \frac{ \tilde{h}_i^2}{ \sigma_i^2 }
    h2 = 0.0

    # iterate over conditions
    for sim_i, data_i, sigma_i, mask_i in zip(
        sim, data, sigma, mask, strict=True
    ):
        if mask_i.max(initial=False) is False:
            continue
        # extract relevant values
        sim_x = sim_i[mask_i]  # \tilde{h}_i
        data_x = data_i[mask_i]  # \bar{y}_i
        sigma_x = sigma_i[mask_i]  # \sigma_i
        # update statistics
        s2 = sigma_x**2  # \sigma_i^2
        h += np.nansum(sim_x / s2)
        recnoise += np.nansum(1 / s2)
        yh += np.nansum((sim_x * data_x) / s2)
        y += np.nansum(data_x / s2)
        h2 += np.nansum((sim_x**2) / s2)

    # numerator and denominator in equation 11, each multiplied by
    #  recnoise = \sum_i \frac{1}{ \sigma_i^2 }
    num = y - (yh * h) / h2
    den = recnoise - (h**2) / h2

    # If simulation is essentially constant, then offset and scaling
    # have the same effect. In this case, we set the offset to a dummy value,
    # and estimate scaling. This is because offset is computed first, by the
    # calculator.
    # NB: this will cause issues if e.g. a simulation with dynamics has values
    #     on the order of the zero check here. An attempt to avoid this
    #     is done here by checking that the numerator is not close to zero.
    if np.isclose(den, 0, atol=1e-14) and not np.isclose(num, 0, atol=1e-4):
        # `get_finite_quotient` will now return the dummy value
        den = 0

    return get_finite_quotient(
        numerator=num,
        denominator=den,
        inner_parameter_type=InnerParameterType.OFFSET,
    )


def apply_offset(
    offset_value: float,
    data: list[np.ndarray],
    mask: list[np.ndarray],
    is_data: bool = True,
):
    """Apply offset to data (in-place).

    Acts on data instead of simulation because of:
    `data = scaling * simulation + offset`. The first step of transforming
    data and simulation so they can be compared is to either multiply
    `simulation` by `scaling`, or subtract `offset` from `data`.
    As we only currently have `compute_optimal_offset_coupled`, not
    `compute_optimal_scaling_coupled`, we need to compute optimal offset first,
    so the first step must act on data.

    NB: as this applies to data in-place, it's important to supply a copy of
    data s.t. the original data is not affected.

    Parameters
    ----------
    offset_value:
        The optimal offset for the masked simulations.
    sim:
        All full (unmasked) data.
    mask:
        The masks that indicate the data subset that corresponds to the
        `offset_value`.
    is_data:
        Whether the data is being offset, or the simulation is. If False, the
        offset is added to the simulation instead of subtracted from the data.
    """
    for i in range(len(data)):
        data[i][mask[i]] += -offset_value if is_data else offset_value


def compute_optimal_sigma(
    data: list[np.ndarray], sim: list[np.ndarray], mask: list[np.ndarray]
) -> float:
    """Compute optimal sigma.

    Compute optimal sigmas for the given measurements and model outputs. See
    https://doi.org/10.1093/bioinformatics/btz581 SI Section 3.2 for the
    derivation.
    """
    # numerator, denominator
    num, den = 0.0, 0.0

    # iterate over conditions
    for sim_i, data_i, mask_i in zip(sim, data, mask, strict=True):
        # extract relevant values
        sim_x = sim_i[mask_i]
        data_x = data_i[mask_i]
        # update statistics
        num += np.nansum((data_x - sim_x) ** 2)
        den += sim_x.size

    if num == 0:
        raise AssertionError()
    # compute optimal value
    return np.sqrt(num / den)


def apply_sigma(
    sigma_value: float, sigma: list[np.ndarray], mask: list[np.ndarray]
):
    """Apply optimal sigma to pre-existing sigma arrays (in-place).

    Parameters
    ----------
    sigma_value:
        The optimal sigma value.
    sigma:
        All full (unmasked) sigmas.
    mask:
        The masks that indicate the sigma subset that corresponds to the
        `sigma_value`.
    """
    for i in range(len(sigma)):
        sigma[i][mask[i]] = sigma_value


def compute_bounded_optimal_scaling_offset_coupled(
    data: list[np.ndarray],
    sim: list[np.ndarray],
    sigma: list[np.ndarray],
    s: InnerParameter,
    b: InnerParameter,
    s_opt_value: float,
    b_opt_value: float,
):
    """Compute optimal scaling and offset of a constrained optimization problem.

    Computes the optimal scaling and offset of a constrained optimization in
    case the unconstrained optimization yields a value outside the bounds.
    We know the optimal solution then lies on the boundary of the bounds.
    In the 2D offset-scaling bounded (rectangular) space, after unconstrained
    optimization, if only one parameter is outside the bounds, then there is
    one active edge (constraint) of the rectangle. We perform optimization on
    this edge that is unconstrained in the other parameter. If this new optimum
    is outside the bounds of the other parameter, the nearest vertex is chosen
    as the optimum. If both parameters are outside the bounds, then there are
    two active edges, which are optimized independently as above, then compared.

    Parameters
    ----------
    data:
        The data.
    sim:
        The simulation.
    sigma:
        The noise parameters.
    s:
        The scaling parameter.
    b:
        The offset parameter.
    s_opt_value:
        The optimal scaling value of the unconstrained problem.
    b_opt_value:
        The optimal offset value of the unconstrained problem.

    Returns
    -------
    The optimal scaling and offset of the constrained problem.
    """
    # Define relevant data and sim
    # Make all non-masked data and sim nan's in the original one
    relevant_data = copy.deepcopy(data)
    relevant_sim = copy.deepcopy(sim)
    for i in range(len(data)):
        relevant_data[i][~s.ixs[i]] = np.nan
        relevant_sim[i][~s.ixs[i]] = np.nan

    # Get relevant data mask
    relevant_data_mask = []
    for i in range(len(data)):
        relevant_data_mask.append(~np.isnan(relevant_data[i]))

    # Get bounds
    s_bounds = s.get_bounds()
    b_bounds = b.get_bounds()

    # Get unsatisfied bounds
    s_unsatisfied = s.get_unsatisfied_bound(s_opt_value)
    b_unsatisfied = b.get_unsatisfied_bound(b_opt_value)

    # If both parameters are unsatisfied, we need to check 2
    # unconstrained problems, clip the solutions to the bounds, and
    # choose the one with the lowest objective value
    if s_unsatisfied is not None and b_unsatisfied is not None:
        # Solve the two unconstrained problems
        candidate_points = [
            (
                s_bounds[s_unsatisfied],
                np.clip(
                    compute_optimal_offset(
                        data, sim, sigma, s.ixs, s_bounds[s_unsatisfied]
                    ),
                    b_bounds[LOWER_BOUND],
                    b_bounds[UPPER_BOUND],
                ),
            ),
            (
                np.clip(
                    compute_optimal_scaling(
                        data, sim, sigma, s.ixs, b_bounds[b_unsatisfied]
                    ),
                    s_bounds[LOWER_BOUND],
                    s_bounds[UPPER_BOUND],
                ),
                b_bounds[b_unsatisfied],
            ),
        ]

        # Evaluate the objective function at the candidate points
        candidate_objective_values = [
            compute_nllh(
                data=relevant_data,
                sim=[
                    sim_i * candidate_point[0] + candidate_point[1]
                    for sim_i in relevant_sim
                ],
                sigma=sigma,
                data_mask=relevant_data_mask,
            )
            for candidate_point in candidate_points
        ]
        # The constrained solution is the candidate point with the lowest
        # objective value
        constrained_solution = candidate_points[
            np.argmin(candidate_objective_values)
        ]

    # If only one parameter is unsatisfied, we need to solve a
    # unconstrained problem, clipped to its boundary
    elif s_unsatisfied is not None:
        # Solve the unconstrained problem
        constrained_solution = (
            s_bounds[s_unsatisfied],
            np.clip(
                compute_optimal_offset(
                    data, sim, sigma, s.ixs, s_bounds[s_unsatisfied]
                ),
                b_bounds[LOWER_BOUND],
                b_bounds[UPPER_BOUND],
            ),
        )
    elif b_unsatisfied is not None:
        # Solve the unconstrained problem
        constrained_solution = (
            np.clip(
                compute_optimal_scaling(
                    data, sim, sigma, s.ixs, b_bounds[b_unsatisfied]
                ),
                s_bounds[LOWER_BOUND],
                s_bounds[UPPER_BOUND],
            ),
            b_bounds[b_unsatisfied],
        )

    return constrained_solution


def compute_nllh(
    data: list[np.ndarray],
    sim: list[np.ndarray],
    sigma: list[np.ndarray],
    data_mask: list[np.ndarray],
) -> float:
    """Compute negative log-likelihood.

    Compute negative log-likelihood of the data, given the model outputs and
    sigmas.
    """
    nllh = 0.0
    for data_i, sim_i, sigma_i, data_mask_i in zip(
        data, sim, sigma, data_mask, strict=True
    ):
        # Mask the data, sim and sigma
        data_i = data_i[data_mask_i]
        sim_i = sim_i[data_mask_i]
        sigma_i = sigma_i[data_mask_i]

        # Compute the negative log-likelihood
        nllh += 0.5 * np.nansum(
            np.log(2 * np.pi * sigma_i**2)
        ) + 0.5 * np.nansum((data_i - sim_i) ** 2 / sigma_i**2)

    return nllh


def compute_nllh_gradient_for_condition(
    data: np.ndarray,
    sim: np.ndarray,
    sigma: np.ndarray,
    ssim: np.ndarray,
    ssigma: np.ndarray,
):
    """Compute gradient of negative the log-likelihood function for a condition.

    Compute gradient of negative log-likelihood function with respect to
    outer optimization parameters for a condition, given the model outputs,
    relevant data and sigmas.
    """
    return np.nansum(
        np.multiply(
            ssigma,
            ((np.full(data.shape, 1) - (data - sim) ** 2 / sigma**2) / sigma),
        ),
        axis=(1, 2),
    ) + np.nansum(
        np.multiply(ssim, (sim - data) / sigma**2),
        axis=(1, 2),
    )



"""Engines with multi-process parallelization."""

import logging
import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Any

import cloudpickle as pickle

from ..util import tqdm
from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(pickled_task) -> Any:
    """Unpickle and execute task.

    Parameters
    ----------
    pickled_task:
        A pickled Task object to execute.

    Returns
    -------
    The result of executing the task.
    """
    task = pickle.loads(pickled_task)
    return task.execute()


class MultiProcessEngine(Engine):
    """
    Parallelize the task execution using multiprocessing.

    Parameters
    ----------
    n_procs:
        The maximum number of processes to use in parallel.
        Defaults to the number of CPUs available on the system according to
        `os.cpu_count()`.
        The effectively used number of processes will be the minimum of
        `n_procs` and the number of tasks submitted. Defaults to ``None``.
    method:
        Start method, any of "fork", "spawn", "forkserver", or None,
        giving the system specific default context. Defaults to ``None``.
    """

    def __init__(
        self,
        n_procs: int | None = None,
        method: str | None = None,
    ):
        super().__init__()

        if n_procs is None:
            n_procs = os.cpu_count()
            logger.info(
                f"Engine will use up to {n_procs} processes (= CPU count)."
            )
        self.n_procs: int = n_procs
        self.method: str = method

    def execute(
        self, tasks: list[Task], progress_bar: bool = None
    ) -> list[Any]:
        """Pickle tasks and distribute work over parallel processes.

        Tasks are pickled on-demand as workers become available.

        Parameters
        ----------
        tasks:
            List of :class:`pypesto.engine.Task` to execute.
        progress_bar:
            Whether to display a progress bar.

        Returns
        -------
        A list of results in the same order as the input tasks.
        """
        n_tasks = len(tasks)
        n_procs = min(self.n_procs, n_tasks)
        logger.debug(f"Parallelizing on {n_procs} processes.")

        ctx = multiprocessing.get_context(method=self.method)

        # Use ProcessPoolExecutor for on-demand pickling
        with ProcessPoolExecutor(
            max_workers=n_procs, mp_context=ctx
        ) as executor:
            # Submit tasks and track futures
            future_to_index = {
                executor.submit(work, pickle.dumps(task)): i
                for i, task in enumerate(tasks)
            }

            # Collect results in original order
            results = [None] * n_tasks
            for future in tqdm(
                as_completed(future_to_index),
                total=n_tasks,
                enable=progress_bar,
            ):
                index = future_to_index[future]
                results[index] = future.result()

        return results



"""
Variational inference
======

Find the best variational approximation in a given family to a distribution from which we can sample.
"""

from .pymc import PymcVariational
from .variational_inference import variational_fit



"""Manage all components of a pyPESTO model selection problem."""

import warnings
from typing import Any

import petab_select
from petab_select import Model, Models

from .method import MethodCaller
from .model_problem import TYPE_POSTPROCESSOR, ModelProblem  # noqa: F401


class Problem:
    """Handles use of a model selection algorithm.

    Handles model selection. Usage involves initialisation with a model
    specifications file, and then calling the :meth:`select` method to perform
    model selection with a specified algorithm and criterion.

    Attributes
    ----------
    calibrated_models : petab_select.Models
        All calibrated models.
    newly_calibrated_models : petab_select.Models
        All models that were calibrated in the latest iteration of model
        selection.
    method_caller : pypesto.select.method.MethodCaller
        Used to run a single iteration of a model
        selection method.
    model_problem_options : dict[str, typing.Any]
        Keyword arguments, passed to the constructor of
        :class:`ModelProblem`.
    petab_select_problem : petab_select.Problem
        A PEtab Select problem.
    """

    # FIXME rename `best_model` to `selected_model` everywhere

    def __init__(
        self,
        petab_select_problem: petab_select.Problem,
        model_postprocessor: TYPE_POSTPROCESSOR | None = None,
        model_problem_options: dict = None,
    ):
        self.petab_select_problem = petab_select_problem

        self.model_problem_options = {}
        if model_problem_options is not None:
            self.model_problem_options = model_problem_options
        # TODO deprecated
        if model_postprocessor is not None:
            warnings.warn(
                "Specifying `model_postprocessor` directly is deprecated. "
                "Please specify it with `model_problem_options`, e.g. "
                'model_problem_options={"postprocessor": ...}`.',
                DeprecationWarning,
                stacklevel=1,
            )
            self.model_problem_options["postprocessor"] = model_postprocessor

        self.set_state(
            calibrated_models=Models(),
            newly_calibrated_models=Models(),
        )

        # TODO default caller, based on petab_select.Problem
        self.method_caller = None

    def create_method_caller(self, **kwargs) -> MethodCaller:
        """Create a method caller.

        ``kwargs`` are passed to the :class:`MethodCaller` constructor.

        Returns
        -------
        A :class:`MethodCaller` instance.
        """
        kwargs = kwargs.copy()
        model_problem_options = self.model_problem_options | kwargs.pop(
            "model_problem_options", {}
        )

        return MethodCaller(
            petab_select_problem=self.petab_select_problem,
            calibrated_models=self.calibrated_models,
            model_problem_options=model_problem_options,
            **kwargs,
        )

    def set_state(
        self,
        calibrated_models: Models,
        newly_calibrated_models: Models,
    ) -> None:
        """Set the state of the problem.

        See :class:`Problem` attributes for argument documentation.
        """
        self.calibrated_models = calibrated_models
        self.newly_calibrated_models = newly_calibrated_models

    def update_with_newly_calibrated_models(
        self,
        newly_calibrated_models: Models | None = None,
    ) -> None:
        """Update the state of the problem with newly calibrated models.

        Args:
            newly_calibrated_models:
                See attributes of :class:`Problem`.
        """
        self.newly_calibrated_models = newly_calibrated_models
        self.calibrated_models += self.newly_calibrated_models

    def handle_select_kwargs(
        self,
        kwargs: dict[str, Any],
    ):
        """Check keyword arguments to select calls."""
        if "newly_calibrated_models" in kwargs:
            raise ValueError(
                "Please supply `newly_calibrated_models` via "
                "`pypesto.select.Problem.set_state`."
            )
        if "calibrated_models" in kwargs:
            raise ValueError(
                "Please supply `calibrated_models` via "
                "`pypesto.select.Problem.set_state`."
            )

    def select(
        self,
        **kwargs,
    ) -> tuple[Model, Models]:
        """Run a single iteration of a model selection algorithm.

        The result is the selected model for the current run, independent of
        previous selected models.

        ``kwargs`` are passed to the :class:`MethodCaller` constructor.

        Returns
        -------
        A 2-tuple, with the following values from this iteration:

           1. the best model; and
           2. all models.
        """
        self.handle_select_kwargs(kwargs)
        method_caller = self.create_method_caller(**kwargs)
        newly_calibrated_models = method_caller()

        self.update_with_newly_calibrated_models(
            newly_calibrated_models=newly_calibrated_models,
        )

        best_model = petab_select.ui.get_best(
            problem=self.petab_select_problem,
            models=self.newly_calibrated_models,
            criterion=method_caller.criterion,
        )

        return best_model, newly_calibrated_models

    def select_to_completion(
        self,
        **kwargs,
    ) -> Models:
        """Perform model selection until the method terminates.

        ``kwargs`` are passed to the :class:`MethodCaller` constructor.

        Returns
        -------
        All models.
        """
        calibrated_models = Models(problem=self.petab_select_problem)
        self.handle_select_kwargs(kwargs)
        method_caller = self.create_method_caller(**kwargs)

        while True:
            try:
                iteration_calibrated_models = method_caller()
                self.update_with_newly_calibrated_models(
                    newly_calibrated_models=iteration_calibrated_models,
                )
                calibrated_models += iteration_calibrated_models
            except StopIteration:
                break

        return calibrated_models

    # TODO method that automatically generates initial models, for a specific
    # number of starts. TODO parallelise?
    def multistart_select(
        self,
        predecessor_models: Models = None,
        **kwargs,
    ) -> tuple[Model, list[Model]]:
        """Run an algorithm multiple times, with different predecessor models.

        Note that the same method caller is currently shared between all calls.
        This may change when parallelization is implemented, but for now
        ensures that the same model isn't calibrated twice.
        Could also be managed by sharing the same "calibrated_models" object
        (but then the same model could be repeatedly calibrated, if the
        calibrations start before any have stopped).

        ``kwargs`` are passed to the :class:`MethodCaller` constructor.

        Parameters
        ----------
        predecessor_models:
            The models that will be used as initial models. One "model
            selection iteration" will be run for each predecessor model.

        Returns
        -------
        A 2-tuple, with the following values:

           1. the best model; and
           2. the best models (the best model at each iteration).
        """
        self.handle_select_kwargs(kwargs)
        model_lists = []

        for predecessor_model in predecessor_models:
            method_caller = self.create_method_caller(
                **(kwargs | {"predecessor_model": predecessor_model})
            )
            models = method_caller()
            self.calibrated_models += models

            model_lists.append(models)
            method_caller.candidate_space.reset()

        best_model = petab_select.ui.get_best(
            problem=method_caller.petab_select_problem,
            models=[model for models in model_lists for model in models],
            criterion=method_caller.criterion,
        )

        return best_model, model_lists



import logging
import warnings
from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator

from ..C import (
    COLOR,
    TRACE_X_STEPS,
    TRACE_X_TIME,
    TRACE_Y_FVAL,
    TRACE_Y_GRADNORM,
)
from ..history import HistoryBase
from ..result import Result
from .clust_color import assign_colors
from .misc import process_offset_y, process_result_list, process_y_limits
from .reference_points import ReferencePoint, create_references

logger = logging.getLogger(__name__)


def optimizer_history(
    results: Result | list[Result],
    ax: plt.Axes | None = None,
    size: tuple = (18.5, 10.5),
    trace_x: str = TRACE_X_STEPS,
    trace_y: str = TRACE_Y_FVAL,
    scale_y: str = "log10",
    offset_y: float | None = None,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    y_limits: float | list[float] | np.ndarray | None = None,
    start_indices: int | list[int] | None = None,
    reference: ReferencePoint
    | dict
    | list[ReferencePoint]
    | list[dict]
    | None = None,
    legends: str | list[str] | None = None,
) -> plt.Axes:
    """
    Plot history of optimizer.

    Can plot either the history of the cost function or of the gradient
    norm, over either the optimizer steps or the computation time.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    ax:
        Axes object to use.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    trace_x:
        What should be plotted on the x-axis?
        Possibilities: TRACE_X
        Default: TRACE_X_STEPS
    trace_y:
        What should be plotted on the y-axis?
        Possibilities: TRACE_Y_FVAL, TRACE_Y_GRADNORM
        Default: TRACE_Y_FVAl
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    offset_y:
        Offset for the y-axis-values, as these are plotted on a log10-scale
        Will be computed automatically if necessary
    colors:
        Color recognized by matplotlib or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically
    y_limits:
        maximum value to be plotted on the y-axis, or y-limits
    start_indices:
        list of integers specifying the multistart to be plotted or
        int specifying up to which start index should be plotted
    reference:
        List of reference points for optimization results, containing at
        least a function value fval
    legends:
        Labels for line plots, one label per result object

    Returns
    -------
    ax:
        The plot axes.
    """
    if isinstance(start_indices, int):
        start_indices = list(range(start_indices))

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    for j, result in enumerate(results):
        # extract cost function values from result
        vals = get_trace(result, trace_x, trace_y)

        # compute the necessary offset for the y-axis
        (vals, offset_y) = get_vals(
            vals, scale_y, offset_y, trace_y, start_indices
        )

        x_label, y_label = get_labels(trace_x, trace_y, offset_y)

        # call lowlevel plot routine
        ax = optimizer_history_lowlevel(
            vals,
            scale_y=scale_y,
            ax=ax,
            colors=colors[j],
            size=size,
            x_label=x_label,
            y_label=y_label,
            legend_text=legends[j],
        )

    # parse and apply plotting options
    ref = create_references(references=reference)

    # handle options
    ax = handle_options(ax, vals, trace_y, ref, y_limits, offset_y)

    return ax


def optimizer_history_lowlevel(
    vals: list[np.ndarray],
    scale_y: str = "log10",
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    ax: plt.Axes | None = None,
    size: tuple = (18.5, 10.5),
    x_label: str = "Optimizer steps",
    y_label: str = "Objective value",
    legend_text: str | None = None,
) -> plt.Axes:
    """
    Plot optimizer history using list of numpy arrays.

    Parameters
    ----------
    vals:
        list of 2xn-arrays (x_values and y_values of the trace)
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    colors:
        Color recognized by matplotlib or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically
    ax:
        Axes object to use.
    size:
        see waterfall
    x_label:
        label for x-axis
    y_label:
        label for y-axis
    legend_text:
        Label for line plots

    Returns
    -------
    ax:
        The plot axes.
    """
    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # parse input
    fvals = []
    if isinstance(vals, list):
        # convert entries to numpy arrays
        for val in vals:
            # val is already an numpy array
            val = np.asarray(val)
            fvals.append(val[1, -1])
    else:
        # convert to a list of numpy arrays
        vals = np.asarray(vals)
        if vals.shape[0] != 2 or vals.ndim != 2:
            raise ValueError(
                "If numpy array is passed directly to lowlevel "
                "routine of optimizer_history, shape needs to "
                "be 2 x n."
            )
        fvals = [vals[1, -1]]
        vals = [vals]
    n_fvals = len(fvals)

    # assign colors
    # note: this has to happen before sorting
    # to get the same colors in different plots
    colors = assign_colors(fvals, colors)

    # sort
    indices = sorted(range(n_fvals), key=lambda j: fvals[j])

    # plot
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    for j, val in enumerate(vals):
        # collect and parse data
        j_fval = indices[j]
        color = colors[j_fval]
        if j == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None

        # line plots
        if scale_y == "log10":
            ax.semilogy(val[0, :], val[1, :], color=color, label=tmp_legend)
        else:
            ax.plot(val[0, :], val[1, :], color=color, label=tmp_legend)

    # set labels
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title("Optimizer history")
    if legend_text is not None:
        ax.legend()

    return ax


def get_trace(
    result: Result, trace_x: str | None, trace_y: str | None
) -> list[np.ndarray]:
    """
    Get the values of the optimizer trace from the pypesto.Result object.

    Parameters
    ----------
    result: pypesto.Result
        Optimization result obtained by 'optimize.py'.
    trace_x: str, optional
        What should be plotted on the x-axis?
        Possibilities: TRACE_X
        Default: TRACE_X_STEPS
    trace_y: str, optional
        What should be plotted on the y-axis?
        Possibilities: TRACE_Y_FVAL, TRACE_Y_GRADNORM
        Default: TRACE_Y_FVAL

    Returns
    -------
    vals:
        list of (x,y)-values.
    x_label:
        label for x-axis to be plotted later.
    y_label:
        label for y-axis to be plotted later.
    """
    # get data frames
    histories: list[HistoryBase] = result.optimize_result.history

    vals = []

    for history in histories:
        options = history.options
        if trace_y == TRACE_Y_GRADNORM:
            # retrieve gradient trace, if saved
            if not options.trace_record or not options.trace_record_grad:
                raise ValueError("No gradient trace has been recorded.")
            grads = history.get_grad_trace()
            indices = [
                i
                for i, val in enumerate(grads)
                if val is not None and np.isfinite(val).all()
            ]

            grads = np.array([grads[i] for i in indices])

            # Get gradient trace, compute norm
            y_vals = np.linalg.norm(grads, axis=1)

        else:  # trace_y == TRACE_Y_FVAL:
            if not options.trace_record:
                raise ValueError("No function value trace has been recorded.")
            fvals = history.get_fval_trace()
            indices = [
                i
                for i, val in enumerate(fvals)
                if val is not None and np.isfinite(val)
            ]

            y_vals = np.array([fvals[i] for i in indices])

        # retrieve values from dataframe
        if trace_x == TRACE_X_TIME:
            times = np.array(history.get_time_trace())
            x_vals = times[indices]

        else:  # trace_x == TRACE_X_STEPS:
            x_vals = np.array(list(range(len(indices))))

        # if the trace is empty, skip
        if len(x_vals) == 0:
            continue

        # write down values
        vals.append(np.vstack([x_vals, y_vals]))

    return vals


def get_vals(
    vals: list[np.ndarray],
    scale_y: str | None,
    offset_y: float,
    trace_y: str,
    start_indices: Iterable[int],
) -> tuple[list[np.ndarray], float]:
    """
    Postprocess the values of the optimization history.

    Depending on the options set by the user (e.g. scale_y, offset_y,
    start_indices).

    Parameters
    ----------
    vals:
        list of numpy arrays of dimension 2 x len(start_indices)
    scale_y:
        May be logarithmic or linear ('log10' or 'lin')
    offset_y:
        offset for the y-axis, as this is supposed to be in log10-scale
    trace_y:
        What should be plotted on the y-axis
    start_indices:
        list of integers specifying the multi start indices to be plotted

    Returns
    -------
    vals:
        list of numpy arrays
    offset_y:
        offset for the y-axis, if this is supposed to be in log10-scale
    y_label:
        Label for y axis
    """
    # get list of indices
    if start_indices is None:
        start_indices = np.array(range(len(vals)))
    else:
        # check whether list or maximum value
        start_indices = np.array(start_indices)

        # check, whether index set is not too big
        existing_indices = np.array(range(len(vals)))
        start_indices = np.intersect1d(start_indices, existing_indices)

    # reduce values to listed values
    vals = [val for i, val in enumerate(vals) if i in start_indices]

    # get the minimal value which should be plotted
    min_val = np.inf
    for val in vals:
        tmp_min = np.min(val[1, :])
        min_val = np.min([min_val, tmp_min])

    # check, whether offset can be used with this data
    if trace_y == TRACE_Y_FVAL:
        offset_y = process_offset_y(offset_y, scale_y, min_val)
    else:
        offset_y = 0.0

    if offset_y != 0:
        for val in vals:
            val[1, :] += offset_y * np.ones(val[1].shape)

    return vals, offset_y


def get_labels(trace_x: str, trace_y: str, offset_y: float) -> tuple[str, str]:
    """
    Generate labels for x and y axes of the history plot.

    Parameters
    ----------
    trace_x:
        What should be plotted on the x-axis. Possible values: TRACE_X.
    trace_y:
        What should be plotted on the y-axis.
        Possible values: TRACE_Y_FVAL, TRACE_Y_GRADNORM.
    offset_y:
        Offset for the y-axis-values.
    Returns
    -------
    labels for x and y axes
    """
    x_label = ""
    y_label = ""

    if trace_x == TRACE_X_TIME:
        x_label = "Computation time [s]"
    else:
        x_label = "Optimizer steps"

    if trace_y == TRACE_Y_GRADNORM:
        y_label = "Gradient norm"
    else:
        y_label = "Objective value"

    if offset_y != 0:
        y_label = "Offsetted " + y_label.lower()

    return x_label, y_label


def handle_options(
    ax: plt.Axes,
    vals: list[np.ndarray],
    trace_y: str,
    ref: list[ReferencePoint],
    y_limits: float | np.ndarray | None,
    offset_y: float,
) -> plt.Axes:
    """
    Apply post-plotting transformations to the axis object.

    Get the limits for the y-axis, plots the reference points, will do
    more at a later time point.

    Parameters
    ----------
    ref:
        List of reference points for optimization results, containing et
        least a function value fval
    vals:
        list of numpy arrays of size 2 x number of values
    trace_y:
        What should be plotted on the x-axis.
    ax:
        Axes object to use.
    y_limits:
        maximum value to be plotted on the y-axis, or y-limits
    offset_y:
        offset for the y-axis, if this is supposed to be in log10-scale

    Returns
    -------
    ax:
        The plot axes.
    """
    # handle y-limits
    ax = process_y_limits(ax, y_limits)

    if trace_y == TRACE_Y_FVAL:
        # handle reference points
        if len(ref) > 0:
            # plot reference points
            # get length of longest trajectory
            max_len = 0
            for val in vals:
                max_len = np.max([max_len, val[0, -1]])

            for i_ref in ref:
                ax.plot(
                    [0, max_len],
                    [i_ref.fval + offset_y, i_ref.fval + offset_y],
                    "--",
                    color=i_ref.color,
                    label=i_ref.legend,
                )

                # create legend for reference points
                if i_ref.legend is not None:
                    ax.legend()
    else:
        logger.warning(
            f"Reference point is currently only implemented for trace_y == "
            f"{TRACE_Y_FVAL} and will not be plotted for trace_y == {trace_y}."
        )

    return ax


def monotonic_history(
    histories: list[HistoryBase],
) -> tuple[np.ndarray, np.ndarray]:
    """
    Get the overall monotonically decreasing history from multiple histories.

    Parameters
    ----------
    histories:
        List of histories to merge.
        The histories are expected to have time and function value traces,
        and that the function values are monotonically decreasing within each
        history.

    Returns
    -------
    t:
        Time points of overall history.
    fx:
        Objective values of overall history.
    """
    if len(histories) == 0:
        warnings.warn("No histories to process.", stacklevel=2)
        return np.array([], dtype="float"), np.array([], dtype="float")

    for history in histories:
        fvals = history.get_fval_trace()
        if not np.all(fvals[:-1] >= fvals[1:]):
            raise ValueError(
                "Each history is expected to have monotonically "
                "decreasing function values."
            )

    # merge results
    t = np.hstack([history.get_time_trace() for history in histories])
    fx = np.hstack([history.get_fval_trace() for history in histories])
    time_order = np.argsort(t)
    t = t[time_order]
    fx = fx[time_order]

    # get the monotonously decreasing sequence
    monotone = np.where(fx == np.fmin.accumulate(fx))[0]
    t_mono, fx_mono = t[monotone], fx[monotone]

    # remove duplicates
    # convert to np.recarray for multi-level sorting
    # so that we can use np.unique later to remove duplicates
    records = np.rec.fromarrays([t_mono, fx_mono], names="t,fx")
    records.sort(order=["t", "fx"])
    t_mono, unique_idx = np.unique(records.t, return_index=True, sorted=True)
    fx_mono = records.fx[unique_idx]

    # extend from last decrease to last timepoint
    if t_mono[-1] != (t_max := t.max()):
        t_mono = np.append(t_mono, [t_max])
        fx_mono = np.append(fx_mono, [fx_mono.min()])

    return t_mono, fx_mono


def sacess_history(
    histories: list[HistoryBase],
    ax: plt.Axes | None = None,
) -> plt.Axes:
    """Plot `SacessOptimizer` history.

    Plot the history of the best objective values for each
    :class:`pypesto.optimize.ess.sacess.SacessOptimizer`
    worker over computation time as step splot.

    Parameters
    ----------
    histories:
        List of histories from different workers as obtained from
        :attr:`pypesto.optimize.ess.sacess.SacessOptimizer.histories`.
    ax:
        Axes object to use.

    Returns
    -------
    The plot axes. `ax` or a new axes if `ax` was `None`.
    """
    ax = ax or plt.subplot()
    if len(histories) == 0:
        warnings.warn("No histories to plot.", stacklevel=2)

    # plot overall minimum
    t_overall, fx_overall = monotonic_history(histories)
    ax.step(
        t_overall,
        fx_overall,
        linestyle="dotted",
        color="grey",
        where="post",
        label="overall",
        alpha=0.8,
    )

    # plot steps of individual workers
    for worker_idx, history in enumerate(histories):
        x, y = history.get_time_trace(), history.get_fval_trace()
        if len(x) == 0:
            warnings.warn(f"No trace for worker #{worker_idx}.", stacklevel=2)
            continue
        # extend from last decrease to last timepoint
        x = np.append(x, [np.max(t_overall)])
        y = np.append(y, [np.min(y)])
        lines = ax.step(
            x, y, ".-", where="post", label=f"worker {worker_idx}", alpha=0.8
        )
        # Plot last point without marker, unless we actually had an improvement there.
        # The time point of the overall last improvement is appended to all histories,
        # even if redundant, so we can just skip the marker for the last point.
        for line in lines:
            line.set_markevery([True] * (len(x) - 1) + [False])

    ax.legend()
    ax.set_xlabel("time (s)")
    ax.set_ylabel("fval")
    ax.set_title("SacessOptimizer convergence")
    return ax



"""Importer for PEtab problems using RoadRunner.

Creates from a PEtab problem a roadrunner model, a roadrunner objective or a
pypesto problem with a roadrunner objective. The actual form of the likelihood
depends on the noise model specified in the provided PEtab problem.
"""

from __future__ import annotations

import logging
import numbers
import re
import warnings
from collections.abc import Iterable
from pathlib import Path
from typing import Any

try:
    import petab.v1 as petab
    from petab.v1.C import (
        OBSERVABLE_FORMULA,
        PREEQUILIBRATION_CONDITION_ID,
        SIMULATION_CONDITION_ID,
    )
    from petab.v1.models.sbml_model import SbmlModel
    from petab.v1.parameter_mapping import ParMappingDictQuadruple
except ImportError:
    petab = None

import pypesto.C

from ...problem import Problem
from ...startpoint import StartpointMethod
from ..aggregated import AggregatedObjective
from ..priors import NegLogParameterPriors, get_parameter_prior_dict
from .road_runner import RoadRunnerObjective
from .roadrunner_calculator import RoadRunnerCalculator
from .utils import ExpData

try:
    import libsbml
    import roadrunner
except ImportError:
    roadrunner = None
    libsbml = None

logger = logging.getLogger(__name__)


class PetabImporterRR:
    """
    Importer for PEtab problems using RoadRunner.

    Create a :class:`roadrunner.RoadRunner` instance,
    a :class:`pypesto.objective.RoadRunnerObjective` or a
    :class:`pypesto.problem.Problem` from PEtab files. The actual
    form of the likelihood depends on the noise model specified in the provided PEtab problem.
    For more information, see the
    `PEtab documentation <https://petab.readthedocs.io/en/latest/documentation_data_format.html#noise-distributions>`_.
    """  # noqa

    def __init__(
        self, petab_problem: petab.Problem, validate_petab: bool = True
    ):
        """Initialize importer.

        Parameters
        ----------
        petab_problem:
            Managing access to the model and data.
        validate_petab:
            Flag indicating if the PEtab problem shall be validated.
        """
        warnings.warn(
            "The RoadRunner importer is deprecated and will be removed in "
            "future versions. Please use the generic PetabImporter instead "
            "with `simulator_type='roadrunner'`. Everything else will stay "
            "same.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.petab_problem = petab_problem
        if validate_petab:
            if petab.lint_problem(petab_problem):
                raise ValueError("Invalid PEtab problem.")
        self.rr = roadrunner.RoadRunner()

    @staticmethod
    def from_yaml(yaml_config: Path | str) -> PetabImporterRR:
        """Simplified constructor using a petab yaml file."""
        petab_problem = petab.Problem.from_yaml(yaml_config)

        return PetabImporterRR(petab_problem=petab_problem)

    def _check_noise_formulae(
        self,
        edatas: list[ExpData] | None = None,
        parameter_mapping: list[ParMappingDictQuadruple] | None = None,
    ):
        """Check if the noise formulae are valid.

        Currently, only static values or singular parameters are supported.
        Complex formulae are not supported.
        """
        # check that parameter mapping is available
        if parameter_mapping is None:
            parameter_mapping = self.create_parameter_mapping()
        # check that edatas are available
        if edatas is None:
            edatas = self.create_edatas()
        # save formulae that need to be changed
        to_change = []
        # check that noise formulae are valid
        for i_edata, (edata, par_map) in enumerate(
            zip(edatas, parameter_mapping, strict=True)
        ):
            for j_formula, noise_formula in enumerate(edata.noise_formulae):
                # constant values are allowed
                if isinstance(noise_formula, numbers.Number):
                    continue
                # single parameters are allowed
                if noise_formula in par_map[1].keys():
                    continue
                # extract the observable name via regex pattern
                pattern = r"noiseParameter1_(.*?)($|\s)"
                observable_name = re.search(pattern, noise_formula).group(1)
                to_change.append((i_edata, j_formula, observable_name))
        # change formulae
        formulae_changed = []
        for i_edata, j_formula, obs_name in to_change:
            # assign new parameter, formula in RR and parameter into mapping
            original_formula = edatas[i_edata].noise_formulae[j_formula]
            edatas[i_edata].noise_formulae[j_formula] = (
                f"noiseFormula_{obs_name}"
            )
            # different conditions will have the same noise formula
            if (obs_name, original_formula) not in formulae_changed:
                self.rr.addParameter(f"noiseFormula_{obs_name}", 0.0, False)
                self.rr.addAssignmentRule(
                    f"noiseFormula_{obs_name}",
                    original_formula,
                    forceRegenerate=False,
                )
                self.rr.regenerateModel()
                formulae_changed.append((obs_name, original_formula))

    def _write_observables_to_model(self):
        """Write observables of petab problem to the model."""
        # add all observables as species
        for obs_id in self.petab_problem.observable_df.index:
            self.rr.addParameter(obs_id, 0.0, False)
        # extract all parameters from observable formulas
        parameters = petab.get_output_parameters(
            self.petab_problem.observable_df,
            self.petab_problem.model,
            noise=True,
            observables=True,
        )
        # add all parameters to the model
        for param_id in parameters:
            self.rr.addParameter(param_id, 0.0, False)
        formulae = self.petab_problem.observable_df[
            OBSERVABLE_FORMULA
        ].to_dict()

        # add all observable formulas as assignment rules
        for obs_id, formula in formulae.items():
            self.rr.addAssignmentRule(obs_id, formula, forceRegenerate=False)

        # regenerate model to apply changes
        self.rr.regenerateModel()

    def create_edatas(self) -> list[ExpData]:
        """Create a List of :class:`ExpData` objects from the PEtab problem."""
        # Create Dataframes per condition
        return ExpData.from_petab_problem(self.petab_problem)

    def fill_model(self):
        """Fill the RoadRunner model inplace from the PEtab problem.

        Parameters
        ----------
        return_model:
            Flag indicating if the model should be returned.
        """
        if not isinstance(self.petab_problem.model, SbmlModel):
            raise ValueError(
                "The model is not an SBML model. Using "
                "RoadRunner as simulator requires an SBML model."
            )  # TODO: add Pysb support
        if self.petab_problem.model.sbml_document:
            sbml_document = self.petab_problem.model.sbml_document
        elif self.petab_problem.model.sbml_model:
            sbml_document = (
                self.petab_problem.model.sbml_model.getSBMLDocument()
            )
        else:
            raise ValueError("No SBML model found.")
        sbml_writer = libsbml.SBMLWriter()
        sbml_string = sbml_writer.writeSBMLToString(sbml_document)
        self.rr.load(sbml_string)
        self._write_observables_to_model()

    def create_parameter_mapping(self):
        """Create a parameter mapping from the PEtab problem."""
        simulation_conditions = (
            self.petab_problem.get_simulation_conditions_from_measurement_df()
        )
        mapping = petab.get_optimization_to_simulation_parameter_mapping(
            condition_df=self.petab_problem.condition_df,
            measurement_df=self.petab_problem.measurement_df,
            parameter_df=self.petab_problem.parameter_df,
            observable_df=self.petab_problem.observable_df,
            model=self.petab_problem.model,
        )
        # check whether any species in the condition table are assigned
        species = self.rr.model.getFloatingSpeciesIds()
        # overrides in parameter table are handled already
        overrides = [
            specie
            for specie in species
            if specie in self.petab_problem.condition_df.columns
        ]
        if not overrides:
            return mapping
        for (_, condition), mapping_per_condition in zip(
            simulation_conditions.iterrows(), mapping, strict=True
        ):
            for override in overrides:
                preeq_id = condition.get(PREEQUILIBRATION_CONDITION_ID)
                sim_id = condition.get(SIMULATION_CONDITION_ID)
                if preeq_id:
                    parameter_id_or_value = (
                        self.petab_problem.condition_df.loc[preeq_id, override]
                    )
                    mapping_per_condition[0][override] = parameter_id_or_value
                    if isinstance(parameter_id_or_value, str):
                        mapping_per_condition[2][override] = (
                            self.petab_problem.parameter_df.loc[
                                parameter_id_or_value, petab.PARAMETER_SCALE
                            ]
                        )
                    elif isinstance(parameter_id_or_value, numbers.Number):
                        mapping_per_condition[2][override] = pypesto.C.LIN
                    else:
                        raise ValueError(
                            "The parameter value in the condition table "
                            "is not a number or a parameter ID."
                        )
                if sim_id:
                    parameter_id_or_value = (
                        self.petab_problem.condition_df.loc[sim_id, override]
                    )
                    mapping_per_condition[1][override] = parameter_id_or_value
                    if isinstance(parameter_id_or_value, str):
                        mapping_per_condition[3][override] = (
                            self.petab_problem.parameter_df.loc[
                                parameter_id_or_value, petab.PARAMETER_SCALE
                            ]
                        )
                    elif isinstance(parameter_id_or_value, numbers.Number):
                        mapping_per_condition[3][override] = pypesto.C.LIN
                    else:
                        raise ValueError(
                            "The parameter value in the condition table "
                            "is not a number or a parameter ID."
                        )
        return mapping

    def create_objective(
        self,
        rr: roadrunner.RoadRunner | None = None,
        edatas: ExpData | None = None,
    ) -> RoadRunnerObjective:
        """Create a :class:`pypesto.objective.RoadRunnerObjective`.

        Parameters
        ----------
        rr:
            RoadRunner instance.
        edatas:
            ExpData object.
        """
        roadrunner_instance = rr
        if roadrunner_instance is None:
            roadrunner_instance = self.rr
            self.fill_model()
        if edatas is None:
            edatas = self.create_edatas()

        parameter_mapping = self.create_parameter_mapping()

        # get x_names
        x_names = self.petab_problem.get_x_ids()

        calculator = RoadRunnerCalculator()

        # run the check for noise formulae
        self._check_noise_formulae(edatas, parameter_mapping)

        return RoadRunnerObjective(
            rr=roadrunner_instance,
            edatas=edatas,
            parameter_mapping=parameter_mapping,
            petab_problem=self.petab_problem,
            calculator=calculator,
            x_names=x_names,
            x_ids=x_names,
        )

    def create_prior(self) -> NegLogParameterPriors | None:
        """
        Create a prior from the parameter table.

        Returns None, if no priors are defined.
        """
        prior_list = []

        if petab.OBJECTIVE_PRIOR_TYPE not in self.petab_problem.parameter_df:
            return None

        for i, x_id in enumerate(self.petab_problem.x_ids):
            prior_type_entry = self.petab_problem.parameter_df.loc[
                x_id, petab.OBJECTIVE_PRIOR_TYPE
            ]

            if (
                isinstance(prior_type_entry, str)
                and prior_type_entry != petab.PARAMETER_SCALE_UNIFORM
            ):
                # check if parameter for which prior is defined is a fixed parameter
                if x_id in self.petab_problem.x_fixed_ids:
                    logger.warning(
                        f"Parameter {x_id} is marked as fixed but has a "
                        f"prior defined. This might be unintended."
                    )

                prior_params = [
                    float(param)
                    for param in self.petab_problem.parameter_df.loc[
                        x_id, petab.OBJECTIVE_PRIOR_PARAMETERS
                    ].split(";")
                ]

                scale = self.petab_problem.parameter_df.loc[
                    x_id, petab.PARAMETER_SCALE
                ]

                prior_list.append(
                    get_parameter_prior_dict(
                        i, prior_type_entry, prior_params, scale
                    )
                )
        return NegLogParameterPriors(prior_list)

    def create_startpoint_method(self, **kwargs) -> StartpointMethod:
        """Create a startpoint method.

        Parameters
        ----------
        **kwargs:
            Additional keyword arguments passed on to
            :meth:`pypesto.startpoint.FunctionStartpoints.__init__`.
        """
        from ...petab.util import PetabStartpoints

        return PetabStartpoints(petab_problem=self.petab_problem, **kwargs)

    def create_problem(
        self,
        objective: RoadRunnerObjective | None = None,
        x_guesses: Iterable[float] | None = None,
        problem_kwargs: dict[str, Any] | None = None,
        startpoint_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ) -> Problem:
        """Create a :class:`pypesto.problem.Problem`.

        Parameters
        ----------
        objective:
            Objective as created by :meth:`create_objective`.
        x_guesses:
            Guesses for the parameter values, shape (g, dim), where g denotes
            the number of guesses. These are used as start points in the
            optimization.
        problem_kwargs:
            Passed to :meth:`pypesto.problem.Problem.__init__`.
        startpoint_kwargs:
            Keyword arguments forwarded to
            :meth:`PetabImporter.create_startpoint_method`.
        **kwargs:
            Additional key word arguments passed on to the objective,
            if not provided.

        Returns
        -------
        A :class:`pypesto.problem.Problem` instance.
        """
        if objective is None:
            objective = self.create_objective(**kwargs)

        x_fixed_indices = self.petab_problem.x_fixed_indices
        x_fixed_vals = self.petab_problem.x_nominal_fixed_scaled
        x_ids = self.petab_problem.x_ids
        lb = self.petab_problem.lb_scaled
        ub = self.petab_problem.ub_scaled

        x_scales = [
            self.petab_problem.parameter_df.loc[x_id, petab.PARAMETER_SCALE]
            for x_id in x_ids
        ]

        if problem_kwargs is None:
            problem_kwargs = {}

        if startpoint_kwargs is None:
            startpoint_kwargs = {}

        prior = self.create_prior()

        if prior is not None:
            objective = AggregatedObjective([objective, prior])

        problem = Problem(
            objective=objective,
            lb=lb,
            ub=ub,
            x_fixed_indices=x_fixed_indices,
            x_fixed_vals=x_fixed_vals,
            x_guesses=x_guesses,
            x_names=x_ids,
            x_scales=x_scales,
            x_priors_defs=prior,
            startpoint_method=self.create_startpoint_method(
                **startpoint_kwargs
            ),
            copy_objective=False,
            **problem_kwargs,
        )

        return problem
